from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from ...fable_library.int32 import parse
from ...fable_library.option import value as value_3
from ...fable_library.reflection import (TypeInfo, int32_type, string_type, option_type, record_type)
from ...fable_library.reg_exp import (groups, get_item)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.system_text import (StringBuilder__ctor, StringBuilder__Append_Z721C83C5)
from ...fable_library.types import (Record, to_string)
from ...fable_library.util import ignore
from .regex import ActivePatterns__007CRegex_007C__007C

SemVerAux_Pattern: str = "^(?<major>\\d+)(\\.(?<minor>\\d+))?(\\.(?<patch>\\d+))?(-(?<pre>[0-9A-Za-z-\\.]+))?(\\+(?<build>[0-9A-Za-z-\\.]+))?$"

def _expr696() -> TypeInfo:
    return record_type("ARCtrl.Helper.SemVer.SemVer", [], SemVer, lambda: [("Major", int32_type), ("Minor", int32_type), ("Patch", int32_type), ("PreRelease", option_type(string_type)), ("Metadata", option_type(string_type))])


@dataclass(eq = False, repr = False, slots = True)
class SemVer(Record):
    Major: int
    Minor: int
    Patch: int
    PreRelease: str | None
    Metadata: str | None

SemVer_reflection = _expr696

def SemVer_make(major: int, minor: int, patch: int, pre: str | None=None, meta: str | None=None) -> SemVer:
    return SemVer(major, minor, patch, pre, meta)


def SemVer_create_Z55658624(major: int, minor: int, patch: int, pre: str | None=None, meta: str | None=None) -> SemVer:
    return SemVer(major, minor, patch, pre, meta)


def SemVer_tryOfString_Z721C83C5(str_1: str) -> SemVer | None:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C(SemVerAux_Pattern, str_1)
    if active_pattern_result is not None:
        m: Any = active_pattern_result
        g: Any = groups(m)
        return SemVer_create_Z55658624(parse(get_item(g, "major") or "", 511, False, 32), parse(get_item(g, "minor") or "", 511, False, 32), parse(get_item(g, "patch") or "", 511, False, 32), (get_item(g, "pre") or "") if (get_item(g, "pre") is not None) else None, (get_item(g, "build") or "") if (get_item(g, "build") is not None) else None)

    else: 
        return None



def SemVer__AsString(this: SemVer) -> str:
    sb: Any = StringBuilder__ctor()
    ignore(StringBuilder__Append_Z721C83C5(sb, to_text(printf("%i.%i.%i"))(this.Major)(this.Minor)(this.Patch)))
    if this.PreRelease is not None:
        def _arrow697(__unit: None=None, this: Any=this) -> str:
            arg_3: str = value_3(this.PreRelease)
            return to_text(printf("-%s"))(arg_3)

        ignore(StringBuilder__Append_Z721C83C5(sb, _arrow697()))

    if this.Metadata is not None:
        def _arrow698(__unit: None=None, this: Any=this) -> str:
            arg_4: str = value_3(this.Metadata)
            return to_text(printf("+%s"))(arg_4)

        ignore(StringBuilder__Append_Z721C83C5(sb, _arrow698()))

    return to_string(sb)


__all__ = ["SemVerAux_Pattern", "SemVer_reflection", "SemVer_make", "SemVer_create_Z55658624", "SemVer_tryOfString_Z721C83C5", "SemVer__AsString"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...dynamic_obj.dynamic_obj import DynamicObj
from ...dynamic_obj.dyn_obj import (try_get_property_value, set_property, set_optional_property)
from ...fable_library.option import (value as value_1, default_arg)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import (FSharpRef, Array)
from ..ldobject import (LDNode, LDNode_reflection)

def _expr1552() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LabProcess", None, LabProcess, LDNode_reflection())


class LabProcess(LDNode):
    def __init__(self, id: str, name: Any=None, agent: Any=None, object: Any=None, result: Any=None, additional_type: Array[str] | None=None, executes_lab_protocol: Any | None=None, parameter_value: Any | None=None, end_time: Any | None=None, disambiguating_description: Any | None=None) -> None:
        super().__init__(id, ["bioschemas.org/LabProcess"], default_arg(additional_type, []))
        this: FSharpRef[LabProcess] = FSharpRef(None)
        this.contents = self
        self.init_00408: int = 1
        set_property("name", name, this.contents)
        set_property("agent", agent, this.contents)
        set_property("object", object, this.contents)
        set_property("result", result, this.contents)
        set_optional_property("executesLabProtocol", executes_lab_protocol, this.contents)
        set_optional_property("parameterValue", parameter_value, this.contents)
        set_optional_property("endTime", end_time, this.contents)
        set_optional_property("disambiguatingDescription", disambiguating_description, this.contents)

    def GetName(self, __unit: None=None) -> str:
        this: LabProcess = self
        obj: DynamicObj = this
        if try_get_property_value("name", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("name")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "name") + "\' is set on this \'") + "LabProcess") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "name") + "\' set on this \'") + "LabProcess") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_name() -> Callable[[LabProcess], str]:
        def _arrow1548(lp: LabProcess) -> str:
            return lp.GetName()

        return _arrow1548

    def GetAgent(self, __unit: None=None) -> str:
        this: LabProcess = self
        obj: DynamicObj = this
        if try_get_property_value("agent", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("agent")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "agent") + "\' is set on this \'") + "LabProcess") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "agent") + "\' set on this \'") + "LabProcess") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_agent() -> Callable[[LabProcess], str]:
        def _arrow1549(lp: LabProcess) -> str:
            return lp.GetAgent()

        return _arrow1549

    def GetObject(self, __unit: None=None) -> str:
        this: LabProcess = self
        obj: DynamicObj = this
        if try_get_property_value("object", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("object")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "object") + "\' is set on this \'") + "LabProcess") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "object") + "\' set on this \'") + "LabProcess") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_object() -> Callable[[LabProcess], str]:
        def _arrow1550(lp: LabProcess) -> str:
            return lp.GetObject()

        return _arrow1550

    def GetResult(self, __unit: None=None) -> str:
        this: LabProcess = self
        obj: DynamicObj = this
        if try_get_property_value("result", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("result")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "result") + "\' is set on this \'") + "LabProcess") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "result") + "\' set on this \'") + "LabProcess") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_result() -> Callable[[LabProcess], str]:
        def _arrow1551(lp: LabProcess) -> str:
            return lp.GetResult()

        return _arrow1551


LabProcess_reflection = _expr1552

def LabProcess__ctor_Z43A9BC86(id: str, name: Any=None, agent: Any=None, object: Any=None, result: Any=None, additional_type: Array[str] | None=None, executes_lab_protocol: Any | None=None, parameter_value: Any | None=None, end_time: Any | None=None, disambiguating_description: Any | None=None) -> LabProcess:
    return LabProcess(id, name, agent, object, result, additional_type, executes_lab_protocol, parameter_value, end_time, disambiguating_description)


__all__ = ["LabProcess_reflection"]




from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ..fable_library.reflection import (TypeInfo, string_type, bool_type, obj_type, lambda_type, unit_type, record_type)
from ..fable_library.types import Record

def _expr3() -> TypeInfo:
    return record_type("DynamicObj.PropertyHelper", [], PropertyHelper, lambda: [("Name", string_type), ("IsStatic", bool_type), ("IsDynamic", bool_type), ("IsMutable", bool_type), ("IsImmutable", bool_type), ("GetValue", lambda_type(obj_type, obj_type)), ("SetValue", lambda_type(obj_type, lambda_type(obj_type, unit_type))), ("RemoveValue", lambda_type(obj_type, unit_type))])


@dataclass(eq = False, repr = False, slots = True)
class PropertyHelper(Record):
    Name: str
    IsStatic: bool
    IsDynamic: bool
    IsMutable: bool
    IsImmutable: bool
    GetValue: Callable[[Any], Any]
    SetValue: Callable[[Any, Any], None]
    RemoveValue: Callable[[Any], None]

PropertyHelper_reflection = _expr3

__all__ = ["PropertyHelper_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.comment import (Comment, Remark)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.component import Component
from ...arctrl_core.Process.protocol import (Protocol_make, Protocol, Protocol_create_Z414665E7)
from ...arctrl_core.Process.protocol_parameter import ProtocolParameter
from ...arctrl_core.uri import URIModule_fromString
from ...fable_library.list import (of_array, FSharpList, of_seq, empty, length, singleton, initialize, map as map_1, iterate_indexed, iterate, cons, reverse)
from ...fable_library.map_util import add_to_dict
from ...fable_library.option import (map, default_arg)
from ...fable_library.seq2 import List_distinct
from ...fable_library.util import (string_hash, IEnumerable_1, IEnumerator)
from .comment import (Comment_fromString, Comment_toString)
from .conversions import (ProtocolParameter_fromAggregatedStrings, Component_fromAggregatedStrings, Option_fromValueWithDefault, ProtocolParameter_toAggregatedStrings, Component_toAggregatedStrings)
from .sparse_table import (SparseTable_GetEmptyComments_3ECCA699, SparseTable__TryGetValueDefault_5BAE6133, SparseTable__TryGetValue_11FD62A8, SparseTable, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1)

name_label: str = "Name"

protocol_type_label: str = "Type"

type_term_accession_number_label: str = "Type Term Accession Number"

type_term_source_reflabel: str = "Type Term Source REF"

description_label: str = "Description"

uri_label: str = "URI"

version_label: str = "Version"

parameters_name_label: str = "Parameters Name"

parameters_term_accession_number_label: str = "Parameters Term Accession Number"

parameters_term_source_reflabel: str = "Parameters Term Source REF"

components_name_label: str = "Components Name"

components_type_label: str = "Components Type"

components_type_term_accession_number_label: str = "Components Type Term Accession Number"

components_type_term_source_reflabel: str = "Components Type Term Source REF"

labels: FSharpList[str] = of_array([name_label, protocol_type_label, type_term_accession_number_label, type_term_source_reflabel, description_label, uri_label, version_label, parameters_name_label, parameters_term_accession_number_label, parameters_term_source_reflabel, components_name_label, components_type_label, components_type_term_accession_number_label, components_type_term_source_reflabel])

def from_string(name: str | None, protocol_type: str | None, type_term_accession_number: str | None, type_term_source_ref: str | None, description: str | None, uri: str | None, version: str | None, parameters_name: str, parameters_term_accession_number: str, parameters_term_source_ref: str, components_name: str, components_type: str, components_type_term_accession_number: str, components_type_term_source_ref: str, comments: FSharpList[Comment]) -> Protocol:
    protocol_type_1: OntologyAnnotation = OntologyAnnotation.create(protocol_type, type_term_source_ref, type_term_accession_number)
    parameters: FSharpList[ProtocolParameter] = of_seq(ProtocolParameter_fromAggregatedStrings(";", parameters_name, parameters_term_source_ref, parameters_term_accession_number))
    components: FSharpList[Component] = Component_fromAggregatedStrings(";", components_name, components_type, components_type_term_source_ref, components_type_term_accession_number)
    def _arrow3454(s: str, name: Any=name, protocol_type: Any=protocol_type, type_term_accession_number: Any=type_term_accession_number, type_term_source_ref: Any=type_term_source_ref, description: Any=description, uri: Any=uri, version: Any=version, parameters_name: Any=parameters_name, parameters_term_accession_number: Any=parameters_term_accession_number, parameters_term_source_ref: Any=parameters_term_source_ref, components_name: Any=components_name, components_type: Any=components_type, components_type_term_accession_number: Any=components_type_term_accession_number, components_type_term_source_ref: Any=components_type_term_source_ref, comments: Any=comments) -> str:
        return URIModule_fromString(s)

    def _arrow3455(s_1: str, name: Any=name, protocol_type: Any=protocol_type, type_term_accession_number: Any=type_term_accession_number, type_term_source_ref: Any=type_term_source_ref, description: Any=description, uri: Any=uri, version: Any=version, parameters_name: Any=parameters_name, parameters_term_accession_number: Any=parameters_term_accession_number, parameters_term_source_ref: Any=parameters_term_source_ref, components_name: Any=components_name, components_type: Any=components_type, components_type_term_accession_number: Any=components_type_term_accession_number, components_type_term_source_ref: Any=components_type_term_source_ref, comments: Any=comments) -> str:
        return URIModule_fromString(s_1)

    return Protocol_make(None, map(_arrow3454, name), Option_fromValueWithDefault(OntologyAnnotation(), protocol_type_1), description, map(_arrow3455, uri), version, Option_fromValueWithDefault(empty(), parameters), Option_fromValueWithDefault(empty(), components), Option_fromValueWithDefault(empty(), comments))


def from_sparse_table(matrix: SparseTable) -> FSharpList[Protocol]:
    if (length(matrix.CommentKeys) != 0) if (matrix.ColumnCount == 0) else False:
        return singleton(Protocol_create_Z414665E7(None, None, None, None, None, None, None, None, of_seq(SparseTable_GetEmptyComments_3ECCA699(matrix))))

    else: 
        def _arrow3456(i: int, matrix: Any=matrix) -> Protocol:
            def mapping(k: str) -> Comment:
                return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, i)))

            comments_1: FSharpList[Comment] = map_1(mapping, matrix.CommentKeys)
            return from_string(SparseTable__TryGetValue_11FD62A8(matrix, (name_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (protocol_type_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (type_term_accession_number_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (type_term_source_reflabel, i)), SparseTable__TryGetValue_11FD62A8(matrix, (description_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (uri_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (version_label, i)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (parameters_name_label, i)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (parameters_term_accession_number_label, i)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (parameters_term_source_reflabel, i)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (components_name_label, i)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (components_type_label, i)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (components_type_term_accession_number_label, i)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (components_type_term_source_reflabel, i)), comments_1)

        return initialize(matrix.ColumnCount, _arrow3456)



def to_sparse_table(protocols: FSharpList[Protocol]) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, labels, None, length(protocols) + 1)
    comment_keys: FSharpList[str] = empty()
    def action_1(i: int, p: Protocol, protocols: Any=protocols) -> None:
        i_1: int = (i + 1) or 0
        pt_1: dict[str, Any]
        pt: OntologyAnnotation = default_arg(p.ProtocolType, OntologyAnnotation())
        pt_1 = OntologyAnnotation.to_string_object(pt, True)
        p_agg: dict[str, Any] = ProtocolParameter_toAggregatedStrings(";", default_arg(p.Parameters, empty()))
        c_agg: dict[str, Any] = Component_toAggregatedStrings(";", default_arg(p.Components, empty()))
        add_to_dict(matrix.Matrix, (name_label, i_1), default_arg(p.Name, ""))
        add_to_dict(matrix.Matrix, (protocol_type_label, i_1), pt_1["TermName"])
        add_to_dict(matrix.Matrix, (type_term_accession_number_label, i_1), pt_1["TermAccessionNumber"])
        add_to_dict(matrix.Matrix, (type_term_source_reflabel, i_1), pt_1["TermSourceREF"])
        add_to_dict(matrix.Matrix, (description_label, i_1), default_arg(p.Description, ""))
        add_to_dict(matrix.Matrix, (uri_label, i_1), default_arg(p.Uri, ""))
        add_to_dict(matrix.Matrix, (version_label, i_1), default_arg(p.Version, ""))
        add_to_dict(matrix.Matrix, (parameters_name_label, i_1), p_agg["TermNameAgg"])
        add_to_dict(matrix.Matrix, (parameters_term_accession_number_label, i_1), p_agg["TermAccessionNumberAgg"])
        add_to_dict(matrix.Matrix, (parameters_term_source_reflabel, i_1), p_agg["TermSourceREFAgg"])
        add_to_dict(matrix.Matrix, (components_name_label, i_1), c_agg["NameAgg"])
        add_to_dict(matrix.Matrix, (components_type_label, i_1), c_agg["TermNameAgg"])
        add_to_dict(matrix.Matrix, (components_type_term_accession_number_label, i_1), c_agg["TermAccessionNumberAgg"])
        add_to_dict(matrix.Matrix, (components_type_term_source_reflabel, i_1), c_agg["TermSourceREFAgg"])
        match_value: FSharpList[Comment] | None = p.Comments
        if match_value is not None:
            def action(comment: Comment, i: Any=i, p: Any=p) -> None:
                nonlocal comment_keys
                pattern_input: tuple[str, str] = Comment_toString(comment)
                n: str = pattern_input[0]
                comment_keys = cons(n, comment_keys)
                add_to_dict(matrix.Matrix, (n, i_1), pattern_input[1])

            iterate(action, match_value)


    iterate_indexed(action_1, protocols)
    class ObjectExpr3458:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3457(x: str, y: str) -> bool:
                return x == y

            return _arrow3457

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3458())), matrix.ColumnCount)


def from_rows(prefix: str | None, line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], FSharpList[Protocol]]:
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, labels, line_number) if (prefix is None) else SparseTable_FromRows_Z5579EC29(rows, labels, line_number, prefix)
    return (tupled_arg[0], tupled_arg[1], tupled_arg[2], from_sparse_table(tupled_arg[3]))


def to_rows(prefix: str | None, protocols: FSharpList[Protocol]) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    m: SparseTable = to_sparse_table(protocols)
    if prefix is None:
        return SparseTable_ToRows_759CAFC1(m)

    else: 
        return SparseTable_ToRows_759CAFC1(m, prefix)



__all__ = ["name_label", "protocol_type_label", "type_term_accession_number_label", "type_term_source_reflabel", "description_label", "uri_label", "version_label", "parameters_name_label", "parameters_term_accession_number_label", "parameters_term_source_reflabel", "components_name_label", "components_type_label", "components_type_term_accession_number_label", "components_type_term_source_reflabel", "labels", "from_string", "from_sparse_table", "to_sparse_table", "from_rows", "to_rows"]




from __future__ import annotations
from typing import Any
from ..fable_library.double import try_parse as try_parse_1
from ..fable_library.int32 import try_parse
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, int32_type, float64_type, string_type, union_type)
from ..fable_library.string_ import (to_text, printf)
from ..fable_library.types import (FSharpRef, to_string, Array, Union)
from ..fable_library.util import int32_to_string
from .ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)

def _expr708() -> TypeInfo:
    return union_type("ARCtrl.Value", [], Value, lambda: [[("Item", OntologyAnnotation_reflection())], [("Item", int32_type)], [("Item", float64_type)], [("Item", string_type)]])


class Value(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Ontology", "Int", "Float", "Name"]

    @staticmethod
    def from_string(value: str) -> Value:
        match_value: tuple[bool, int]
        out_arg: int = 0
        def _arrow703(__unit: None=None) -> int:
            return out_arg

        def _arrow704(v: int) -> None:
            nonlocal out_arg
            out_arg = v or 0

        match_value = (try_parse(value, 511, False, 32, FSharpRef(_arrow703, _arrow704)), out_arg)
        if match_value[0]:
            return Value(1, match_value[1])

        else: 
            match_value_1: tuple[bool, float]
            out_arg_1: float = 0.0
            def _arrow705(__unit: None=None) -> float:
                return out_arg_1

            def _arrow706(v_2: float) -> None:
                nonlocal out_arg_1
                out_arg_1 = v_2

            match_value_1 = (try_parse_1(value, FSharpRef(_arrow705, _arrow706)), out_arg_1)
            return Value(2, match_value_1[1]) if match_value_1[0] else Value(3, value)


    @staticmethod
    def from_options(value: str | None=None, term_source: str | None=None, term_accesssion: str | None=None) -> Value | None:
        def _arrow707(__unit: None=None) -> Value | None:
            value_1: str = value
            return Value.from_string(value_1)

        return ((None if (term_accesssion is None) else Value(0, OntologyAnnotation.create(default_arg(value, ""), term_source, term_accesssion))) if (term_source is None) else Value(0, OntologyAnnotation.create(default_arg(value, ""), term_source, term_accesssion))) if (value is None) else ((_arrow707() if (term_accesssion is None) else Value(0, OntologyAnnotation.create(default_arg(value, ""), term_source, term_accesssion))) if (term_source is None) else Value(0, OntologyAnnotation.create(default_arg(value, ""), term_source, term_accesssion)))

    @staticmethod
    def to_options(value: Value) -> tuple[str | None, str | None, str | None]:
        if value.tag == 1:
            return (int32_to_string(value.fields[0]), None, None)

        elif value.tag == 2:
            return (to_string(value.fields[0]), None, None)

        elif value.tag == 3:
            return (value.fields[0], None, None)

        else: 
            oa: OntologyAnnotation = value.fields[0]
            return (oa.Name, oa.TermAccessionNumber, oa.TermSourceREF)


    @property
    def Text(self, __unit: None=None) -> str:
        this: Value = self
        if this.tag == 2:
            return to_string(this.fields[0])

        elif this.tag == 1:
            return int32_to_string(this.fields[0])

        elif this.tag == 3:
            return this.fields[0]

        else: 
            return this.fields[0].NameText


    def AsName(self, __unit: None=None) -> str:
        this: Value = self
        if this.tag == 3:
            return this.fields[0]

        else: 
            raise Exception(("Value " + str(this)) + " is not of case name")


    def AsInt(self, __unit: None=None) -> int:
        this: Value = self
        if this.tag == 1:
            return this.fields[0]

        else: 
            raise Exception(("Value " + str(this)) + " is not of case int")


    def AsFloat(self, __unit: None=None) -> float:
        this: Value = self
        if this.tag == 2:
            return this.fields[0]

        else: 
            raise Exception(("Value " + str(this)) + " is not of case float")


    def AsOntology(self, __unit: None=None) -> OntologyAnnotation:
        this: Value = self
        if this.tag == 0:
            return this.fields[0]

        else: 
            raise Exception(("Value " + str(this)) + " is not of case ontology")


    @property
    def IsAnOntology(self, __unit: None=None) -> bool:
        this: Value = self
        return True if (this.tag == 0) else False

    @property
    def IsNumerical(self, __unit: None=None) -> bool:
        this: Value = self
        if (this.tag == 1) or (this.tag == 2):
            return True

        else: 
            return False


    @property
    def IsAnInt(self, __unit: None=None) -> bool:
        this: Value = self
        return True if (this.tag == 1) else False

    @property
    def IsAFloat(self, __unit: None=None) -> bool:
        this: Value = self
        return True if (this.tag == 2) else False

    @property
    def IsAText(self, __unit: None=None) -> bool:
        this: Value = self
        return True if (this.tag == 3) else False

    @staticmethod
    def get_text(v: Value) -> str:
        return v.Text

    def Print(self, __unit: None=None) -> str:
        this: Value = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: Value = self
        if this.tag == 1:
            return to_text(printf("%i"))(this.fields[0])

        elif this.tag == 2:
            return to_text(printf("%f"))(this.fields[0])

        elif this.tag == 3:
            return this.fields[0]

        else: 
            return this.fields[0].NameText



Value_reflection = _expr708

__all__ = ["Value_reflection"]




from __future__ import annotations

from typing import Any

from .reflection import TypeInfo, class_type
from .string_ import format, is_null_or_empty, join, replace, substring
from .types import Array, to_string
from .util import clear, int32_to_string


def _expr3() -> TypeInfo:
    return class_type("System.Text.StringBuilder", None, StringBuilder)


class StringBuilder:
    def __init__(self, value: str, capacity: int) -> None:
        self.buf: Array[str] = []
        if not is_null_or_empty(value):
            (self.buf.append(value))

    def __str__(self, __unit: None = None) -> str:
        _: StringBuilder = self
        return join("", _.buf)


StringBuilder_reflection = _expr3


def StringBuilder__ctor_Z18115A39(value: str, capacity: int) -> StringBuilder:
    return StringBuilder(value, capacity)


def StringBuilder__ctor_Z524259A4(capacity: int) -> StringBuilder:
    return StringBuilder__ctor_Z18115A39("", capacity)


def StringBuilder__ctor_Z721C83C5(value: str) -> StringBuilder:
    return StringBuilder__ctor_Z18115A39(value, 16)


def StringBuilder__ctor(__unit: None = None) -> StringBuilder:
    return StringBuilder__ctor_Z18115A39("", 16)


def StringBuilder__Append_Z721C83C5(x: StringBuilder, s: str) -> StringBuilder:
    (x.buf.append(s))
    return x


def StringBuilder__Append_487EF8FB(x: StringBuilder, s: str, start_index: int, count: int) -> StringBuilder:
    (x.buf.append(substring(s, start_index, count)))
    return x


def StringBuilder__Append_244C7CD6(x: StringBuilder, c: str) -> StringBuilder:
    (x.buf.append(c))
    return x


def StringBuilder__Append_Z524259A4(x: StringBuilder, o: int) -> StringBuilder:
    (x.buf.append(int32_to_string(o)))
    return x


def StringBuilder__Append_5E38073B(x: StringBuilder, o: float) -> StringBuilder:
    (x.buf.append(to_string(o)))
    return x


def StringBuilder__Append_Z1FBCCD16(x: StringBuilder, o: bool) -> StringBuilder:
    (x.buf.append(to_string(o)))
    return x


def StringBuilder__Append_4E60E31B(x: StringBuilder, o: Any = None) -> StringBuilder:
    (x.buf.append(to_string(o)))
    return x


def StringBuilder__Append_Z372E4D23(x: StringBuilder, cs: Array[str]) -> StringBuilder:
    (x.buf.append("".join(cs)))
    return x


def StringBuilder__Append_43A65C09(x: StringBuilder, s: StringBuilder) -> StringBuilder:
    (x.buf.append(to_string(s)))
    return x


def StringBuilder__AppendFormat_433E080(x: StringBuilder, fmt: str, o: Any = None) -> StringBuilder:
    (x.buf.append(format(fmt, o)))
    return x


def StringBuilder__AppendFormat_Z3B30EC65(x: StringBuilder, fmt: str, o1: Any = None, o2: Any = None) -> StringBuilder:
    (x.buf.append(format(fmt, o1, o2)))
    return x


def StringBuilder__AppendFormat_10D165E0(
    x: StringBuilder, fmt: str, o1: Any = None, o2: Any = None, o3: Any = None
) -> StringBuilder:
    (x.buf.append(format(fmt, o1, o2, o3)))
    return x


def StringBuilder__AppendFormat_Z17053F5(x: StringBuilder, fmt: str, arr: Array[Any]) -> StringBuilder:
    (x.buf.append(format(fmt, *arr)))
    return x


def StringBuilder__AppendFormat_Z696D8D1B(
    x: StringBuilder, provider: IFormatProvider, fmt: str, o: Any = None
) -> StringBuilder:
    (x.buf.append(format(provider, fmt, o)))
    return x


def StringBuilder__AppendFormat_26802C9E(
    x: StringBuilder, provider: IFormatProvider, fmt: str, o1: Any = None, o2: Any = None
) -> StringBuilder:
    (x.buf.append(format(provider, fmt, o1, o2)))
    return x


def StringBuilder__AppendFormat_Z471ADCBB(
    x: StringBuilder, provider: IFormatProvider, fmt: str, o1: Any = None, o2: Any = None, o3: Any = None
) -> StringBuilder:
    (x.buf.append(format(provider, fmt, o1, o2, o3)))
    return x


def StringBuilder__AppendFormat_6C2E3E6E(
    x: StringBuilder, provider: IFormatProvider, fmt: str, arr: Array[Any]
) -> StringBuilder:
    (x.buf.append(format(provider, fmt, *arr)))
    return x


def StringBuilder__AppendLine(x: StringBuilder) -> StringBuilder:
    (x.buf.append("\n"))
    return x


def StringBuilder__AppendLine_Z721C83C5(x: StringBuilder, s: str) -> StringBuilder:
    (x.buf.append(s))
    (x.buf.append("\n"))
    return x


def StringBuilder__Clear(x: StringBuilder) -> StringBuilder:
    clear(x.buf)
    return x


def StringBuilder__get_Chars_Z524259A4(x: StringBuilder, index: int) -> str:
    len_1: int = 0
    i: int = -1
    while (len_1 < index) if ((i + 1) < len(x.buf)) else False:
        i = (i + 1) or 0
        len_1 = (len_1 + len(x.buf[i])) or 0
    if True if (True if (index < 0) else (i < 0)) else (i >= len(x.buf)):
        raise Exception("Index was outside the bounds of the array")

    else:
        pos: int = ((len_1 - index) - 1) or 0
        return x.buf[i][pos]


def StringBuilder__set_Chars_413E0D0A(x: StringBuilder, index: int, value: str) -> None:
    len_1: int = 0
    i: int = -1
    while (len_1 < index) if ((i + 1) < len(x.buf)) else False:
        i = (i + 1) or 0
        len_1 = (len_1 + len(x.buf[i])) or 0
    if True if (True if (index < 0) else (i < 0)) else (i >= len(x.buf)):
        raise Exception("Index was outside the bounds of the array")

    else:
        pos: int = ((len_1 - index) - 1) or 0
        x.buf[i] = (x.buf[i][0 : (pos - 1) + 1] + value) + x.buf[i][pos + 1 : len(x.buf[i])]


def StringBuilder__Replace_Z766F94C0(x: StringBuilder, old_value: str, new_value: str) -> StringBuilder:
    for i in range(len(x.buf) - 1, 0 - 1, -1):
        x.buf[i] = replace(x.buf[i], old_value, new_value)
    return x


def StringBuilder__Replace_Z384F8060(x: StringBuilder, old_value: str, new_value: str) -> StringBuilder:
    str_1: str = replace(to_string(x), old_value, new_value)
    return StringBuilder__Append_Z721C83C5(StringBuilder__Clear(x), str_1)


def StringBuilder__get_Length(x: StringBuilder) -> int:
    len_1: int = 0
    for i in range(len(x.buf) - 1, 0 - 1, -1):
        len_1 = (len_1 + len(x.buf[i])) or 0
    return len_1


def StringBuilder__ToString_Z37302880(x: StringBuilder, first_index: int, length: int) -> str:
    return substring(to_string(x), first_index, length)


__all__ = [
    "StringBuilder_reflection",
    "StringBuilder__ctor_Z524259A4",
    "StringBuilder__ctor_Z721C83C5",
    "StringBuilder__ctor",
    "StringBuilder__Append_Z721C83C5",
    "StringBuilder__Append_487EF8FB",
    "StringBuilder__Append_244C7CD6",
    "StringBuilder__Append_Z524259A4",
    "StringBuilder__Append_5E38073B",
    "StringBuilder__Append_Z1FBCCD16",
    "StringBuilder__Append_4E60E31B",
    "StringBuilder__Append_Z372E4D23",
    "StringBuilder__Append_43A65C09",
    "StringBuilder__AppendFormat_433E080",
    "StringBuilder__AppendFormat_Z3B30EC65",
    "StringBuilder__AppendFormat_10D165E0",
    "StringBuilder__AppendFormat_Z17053F5",
    "StringBuilder__AppendFormat_Z696D8D1B",
    "StringBuilder__AppendFormat_26802C9E",
    "StringBuilder__AppendFormat_Z471ADCBB",
    "StringBuilder__AppendFormat_6C2E3E6E",
    "StringBuilder__AppendLine",
    "StringBuilder__AppendLine_Z721C83C5",
    "StringBuilder__Clear",
    "StringBuilder__get_Chars_Z524259A4",
    "StringBuilder__set_Chars_413E0D0A",
    "StringBuilder__Replace_Z766F94C0",
    "StringBuilder__Replace_Z384F8060",
    "StringBuilder__get_Length",
    "StringBuilder__ToString_Z37302880",
]



from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1680() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.MaterialAttribute.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("MaterialAttribute", string_type), ("ArcMaterialAttribute", string_type), ("characteristic_type", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    MaterialAttribute: str
    ArcMaterialAttribute: str
    characteristic_type: str

IContext_reflection = _expr1680

def _arrow1687(__unit: None=None) -> IEncodable:
    class ObjectExpr1681(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1682(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("http://purl.org/nfdi4plants/ontology/")

    class ObjectExpr1683(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:Property")

    class ObjectExpr1684(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("arc:ARC#ARC_00000050")

    class ObjectExpr1685(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("arc:ARC#ARC_00000098")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1681()), ("arc", ObjectExpr1682()), ("MaterialAttribute", ObjectExpr1683()), ("ArcMaterialAttribute", ObjectExpr1684()), ("characteristicType", ObjectExpr1685())])
    class ObjectExpr1686(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_5))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_5.encode_object(arg)

    return ObjectExpr1686()


context_jsonvalue: IEncodable = _arrow1687()

context_str: str = "\r\n{\r\n  \"@context\": {\r\n    \"sdo\": \"http://schema.org/\",\r\n    \"arc\": \"http://purl.org/nfdi4plants/ontology/\",\r\n\r\n    \"MaterialAttribute\": \"sdo:Property\",\r\n    \"ArcMaterialAttribute\": \"arc:ARC#ARC_00000050\",\r\n\r\n    \"characteristicType\": \"arc:ARC#ARC_00000098\"\r\n  }\r\n}\r\n    "

__all__ = ["IContext_reflection", "context_jsonvalue", "context_str"]




from __future__ import annotations
from typing import Any
from ...fable_library.option import (value, map)
from ...fable_library.reflection import (TypeInfo, class_type)
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDRef)
from .formal_parameter import LDFormalParameter

def _expr1411() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDComputerLanguage", None, LDComputerLanguage)


class LDComputerLanguage:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/ComputerLanguage"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def alternate_name() -> str:
        return "http://schema.org/alternateName"

    @staticmethod
    def identifier() -> str:
        return "http://schema.org/identifier"

    @staticmethod
    def url() -> str:
        return "http://schema.org/url"

    @staticmethod
    def same_as() -> str:
        return "http://schema.org/sameAs"

    @staticmethod
    def try_get_name_as_string(cl: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = cl.TryGetPropertyAsSingleton(LDFormalParameter.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(cl: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = cl.TryGetPropertyAsSingleton(LDComputerLanguage.name(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("Property of `name` of object with @id `" + cl.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `name` of object with @id `" + cl.Id) + "`")


    @staticmethod
    def set_name_as_string(cl: LDNode, name: str, context: LDContext | None=None) -> Any:
        return cl.SetProperty(LDComputerLanguage.name(), name, context)

    @staticmethod
    def try_get_alternate_name_as_string(cl: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = cl.TryGetPropertyAsSingleton(LDComputerLanguage.alternate_name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_alternate_name_as_string(cl: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = cl.TryGetPropertyAsSingleton(LDComputerLanguage.alternate_name(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("Property of `alternateName` of object with @id `" + cl.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `alternateName` of object with @id `" + cl.Id) + "`")


    @staticmethod
    def set_alternate_name_as_string(cl: LDNode, alternate_name: str, context: LDContext | None=None) -> Any:
        return cl.SetProperty(LDComputerLanguage.alternate_name(), alternate_name, context)

    @staticmethod
    def try_get_identifier_as_ldref(cl: LDNode, context: LDContext | None=None) -> LDRef | None:
        match_value: Any | None = cl.TryGetPropertyAsSingleton(LDComputerLanguage.identifier(), context)
        (pattern_matching_result, id) = (None, None)
        if match_value is not None:
            if isinstance(value(match_value), LDRef):
                pattern_matching_result = 0
                id = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return id

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_identifier_as_ldref(cl: LDNode, context: LDContext | None=None) -> LDRef:
        match_value: Any | None = cl.TryGetPropertyAsSingleton(LDComputerLanguage.identifier(), context)
        if match_value is not None:
            if isinstance(value(match_value), LDRef):
                id: LDRef = value(match_value)
                return id

            else: 
                raise Exception(("Property of `identifier` of object with @id `" + cl.Id) + "` was not a LDRef")


        else: 
            raise Exception(("Could not access property `identifier` of object with @id `" + cl.Id) + "`")


    @staticmethod
    def set_identifier_as_ldref(cl: LDNode, identifier: LDRef, context: LDContext | None=None) -> Any:
        return cl.SetProperty(LDComputerLanguage.identifier(), identifier, context)

    @staticmethod
    def try_get_url_as_ldref(cl: LDNode, context: LDContext | None=None) -> LDRef | None:
        match_value: Any | None = cl.TryGetPropertyAsSingleton(LDComputerLanguage.url(), context)
        (pattern_matching_result, url) = (None, None)
        if match_value is not None:
            if isinstance(value(match_value), LDRef):
                pattern_matching_result = 0
                url = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return url

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_url_as_ldref(cl: LDNode, context: LDContext | None=None) -> LDRef:
        match_value: Any | None = cl.TryGetPropertyAsSingleton(LDComputerLanguage.url(), context)
        if match_value is not None:
            if isinstance(value(match_value), LDRef):
                url: LDRef = value(match_value)
                return url

            else: 
                raise Exception(("Property of `url` of object with @id `" + cl.Id) + "` was not a LDRef")


        else: 
            raise Exception(("Could not access property `url` of object with @id `" + cl.Id) + "`")


    @staticmethod
    def set_url_as_ldref(cl: LDNode, url: LDRef, context: LDContext | None=None) -> Any:
        return cl.SetProperty(LDComputerLanguage.url(), url, context)

    @staticmethod
    def validate(cl: LDNode, context: LDContext | None=None) -> bool:
        return cl.HasType(LDComputerLanguage.schema_type(), context)

    @staticmethod
    def validate_cwl(cl: LDNode, context: LDContext | None=None) -> bool:
        return (cl.Id == "https://w3id.org/workflowhub/workflow-ro-crate#cwl") if LDComputerLanguage.validate(cl, context) else False

    @staticmethod
    def validate_galaxy(cl: LDNode, context: LDContext | None=None) -> bool:
        return (cl.Id == "https://w3id.org/workflowhub/workflow-ro-crate#galaxy") if LDComputerLanguage.validate(cl, context) else False

    @staticmethod
    def validate_knime(cl: LDNode, context: LDContext | None=None) -> bool:
        return (cl.Id == "https://w3id.org/workflowhub/workflow-ro-crate#knime") if LDComputerLanguage.validate(cl, context) else False

    @staticmethod
    def validate_nextflow(cl: LDNode, context: LDContext | None=None) -> bool:
        return (cl.Id == "https://w3id.org/workflowhub/workflow-ro-crate#nextflow") if LDComputerLanguage.validate(cl, context) else False

    @staticmethod
    def validate_snakemake(cl: LDNode, context: LDContext | None=None) -> bool:
        return (cl.Id == "https://w3id.org/workflowhub/workflow-ro-crate#snakemake") if LDComputerLanguage.validate(cl, context) else False

    @staticmethod
    def create(id: str, name: str | None=None, alternate_name: str | None=None, identifier: str | None=None, url: str | None=None, context: LDContext | None=None) -> LDNode:
        cl: LDNode = LDNode(id, [LDComputerLanguage.schema_type()], None, context)
        identifier_1: LDRef | None = map(LDRef, identifier)
        url_1: LDRef | None = map(LDRef, url)
        cl.SetOptionalProperty(LDComputerLanguage.name(), name, context)
        cl.SetOptionalProperty(LDComputerLanguage.alternate_name(), alternate_name, context)
        cl.SetOptionalProperty(LDComputerLanguage.identifier(), identifier_1, context)
        cl.SetOptionalProperty(LDComputerLanguage.url(), url_1, context)
        return cl

    @staticmethod
    def create_cwl(context: LDContext | None=None) -> LDNode:
        return LDComputerLanguage.create("https://w3id.org/workflowhub/workflow-ro-crate#cwl", "Common Workflow Language", "CWL", "https://w3id.org/cwl/v1.2/", "https://www.commonwl.org/", context)

    @staticmethod
    def create_galaxy(context: LDContext | None=None) -> LDNode:
        return LDComputerLanguage.create("https://w3id.org/workflowhub/workflow-ro-crate#galaxy", "Galaxy", None, "https://galaxyproject.org/", "https://galaxyproject.org/", context)

    @staticmethod
    def create_knime(context: LDContext | None=None) -> LDNode:
        return LDComputerLanguage.create("https://w3id.org/workflowhub/workflow-ro-crate#knime", "KNIME", None, "https://www.knime.com/", "https://www.knime.com/", context)

    @staticmethod
    def create_nextflow(context: LDContext | None=None) -> LDNode:
        return LDComputerLanguage.create("https://w3id.org/workflowhub/workflow-ro-crate#nextflow", "Nextflow", None, "https://www.nextflow.io/", "https://www.nextflow.io/", context)

    @staticmethod
    def create_snakemake(context: LDContext | None=None) -> LDNode:
        return LDComputerLanguage.create("https://w3id.org/workflowhub/workflow-ro-crate#snakemake", "Snakemake", None, "https://doi.org/10.1093/bioinformatics/bts480", "https://snakemake.readthedocs.io", context)


LDComputerLanguage_reflection = _expr1411

__all__ = ["LDComputerLanguage_reflection"]




from __future__ import annotations

from datetime import datetime, timedelta, timezone

from . import time_span
from .time_span import TimeSpan
from .types import FSharpRef


def timedelta_total_microseconds(td: timedelta) -> int:
    # timedelta doesn't expose total_microseconds
    # so we need to calculate it ourselves
    return td.days * (24 * 3600) + td.seconds * 10**6 + td.microseconds


def add(d: datetime, ts: timedelta) -> datetime:
    return d + ts


def parse(string: str, detectUTC: bool = False) -> datetime:
    from dateutil import parser

    return parser.parse(string)


def try_parse(string: str, style: int, unsigned: bool, bitsize: int, defValue: FSharpRef[datetime]) -> bool:
    try:
        defValue.contents = parse(string)
        return True
    except Exception:
        return False


def create(
    year: int,
    month: int,
    day: int,
    h: int,
    m: int,
    s: int,
    ms: int | TimeSpan,
    offset: TimeSpan | None = None,
) -> datetime:
    python_offset: timedelta | None = None
    if isinstance(ms, TimeSpan):
        python_offset = timedelta(microseconds=time_span.total_microseconds(ms))
        ms = 0

    if python_offset is None:
        return datetime(year, month, day, h, m, s, ms)

    tzinfo = timezone(python_offset)
    return datetime(year, month, day, h, m, s, ms, tzinfo=tzinfo)


def now() -> datetime:
    return datetime.now()


def utc_now() -> datetime:
    return datetime.utcnow()


def op_addition(x: datetime, y: timedelta) -> datetime:
    return x + y


def op_subtraction(x: datetime, y: datetime | TimeSpan) -> datetime | TimeSpan:
    if isinstance(y, TimeSpan):
        return x - timedelta(microseconds=time_span.total_microseconds(y))

    return time_span.create(0, 0, 0, 0, 0, timedelta_total_microseconds(x - y))


def min_value() -> datetime:
    return datetime.min


__all__ = ["now", "op_addition", "op_subtraction", "parse", "try_parse", "utc_now"]



from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.list import (FSharpList, is_empty, reduce)
from ..fable_library.map_util import add_to_dict
from ..fable_library.mutable_map import Dictionary
from ..fable_library.option import (map, value as value_2, default_arg)
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (filter, choose, map as map_1, iterate, sort_by, to_list, delay, enumerate_while, singleton)
from ..fable_library.util import (IEnumerable_1, safe_hash, compare_primitives, is_iterable, get_enumerator, IEnumerator, equals, structural_hash, dispose, is_array_like)
from .fable_py import (try_get_dynamic_property_helper, get_property_helpers)
from .hash_codes import (box_hash_key_val_seq_by, hash_1 as hash_1_1, merge_hashes)
from .property_helper import PropertyHelper
from .reflection_utils import try_get_static_property_info

def _expr59() -> TypeInfo:
    return class_type("DynamicObj.DynamicObj", None, DynamicObj)


class DynamicObj:
    def __init__(self, __unit: None=None) -> None:
        self.properties: Any = dict([])

    @property
    def Properties(self, __unit: None=None) -> Any:
        this: DynamicObj = self
        return this.properties

    @Properties.setter
    def Properties(self, value: Any) -> None:
        this: DynamicObj = self
        this.properties = value

    @staticmethod
    def of_dict(dynamic_properties: Any) -> DynamicObj:
        obj: DynamicObj = DynamicObj()
        obj.Properties = dynamic_properties
        return obj

    def TryGetStaticPropertyHelper(self, property_name: str) -> PropertyHelper | None:
        this: DynamicObj = self
        return try_get_static_property_info(this, property_name)

    def TryGetDynamicPropertyHelper(self, property_name: str) -> PropertyHelper | None:
        this: DynamicObj = self
        return try_get_dynamic_property_helper(this, property_name)

    def TryGetPropertyHelper(self, property_name: str) -> PropertyHelper | None:
        this: DynamicObj = self
        match_value: PropertyHelper | None = this.TryGetStaticPropertyHelper(property_name)
        return this.TryGetDynamicPropertyHelper(property_name) if (match_value is None) else match_value

    def TryGetPropertyValue(self, property_name: str) -> Any | None:
        this: DynamicObj = self
        def mapping(pi: PropertyHelper) -> Any:
            return pi.GetValue(this)

        return map(mapping, this.TryGetPropertyHelper(property_name))

    def GetPropertyValue(self, property_name: str) -> Any:
        this: DynamicObj = self
        match_value: Any | None = this.TryGetPropertyValue(property_name)
        if match_value is None:
            raise Exception(("No dynamic or static property \"" + property_name) + "\" does exist on object.")

        else: 
            return value_2(match_value)


    def SetProperty(self, property_name: str, property_value: Any=None) -> None:
        this: DynamicObj = self
        match_value: PropertyHelper | None = this.TryGetStaticPropertyHelper(property_name)
        if match_value is None:
            setattr(this,property_name,property_value)

        else: 
            pi: PropertyHelper = match_value
            if pi.IsMutable:
                pi.SetValue(this, property_value)

            else: 
                raise Exception(("Cannot set value for static, immutable property \"" + property_name) + "\"")



    def RemoveProperty(self, property_name: str) -> bool:
        this: DynamicObj = self
        match_value: PropertyHelper | None = this.TryGetPropertyHelper(property_name)
        if match_value is None:
            return False

        elif match_value.IsMutable:
            pi_1: PropertyHelper = match_value
            pi_1.RemoveValue(this)
            return True

        else: 
            raise Exception(("Cannot remove value for static, immutable property \"" + property_name) + "\"")


    def GetPropertyHelpers(self, include_instance_properties: bool) -> IEnumerable_1[PropertyHelper]:
        this: DynamicObj = self
        def predicate_1(p: PropertyHelper) -> bool:
            return p.Name.lower() != "properties"

        def predicate(pd: PropertyHelper) -> bool:
            if include_instance_properties:
                return True

            else: 
                return pd.IsDynamic


        return filter(predicate_1, filter(predicate, get_property_helpers(this)))

    def GetProperties(self, include_instance_properties: bool) -> IEnumerable_1[Any]:
        this: DynamicObj = self
        def chooser(kv: PropertyHelper) -> Any | None:
            if kv.Name != "properties":
                return (kv.Name, kv.GetValue(this))

            else: 
                return None


        return choose(chooser, this.GetPropertyHelpers(include_instance_properties))

    def GetPropertyNames(self, include_instance_properties: bool) -> IEnumerable_1[str]:
        this: DynamicObj = self
        def mapping(kv: Any) -> str:
            return kv[0]

        return map_1(mapping, this.GetProperties(include_instance_properties))

    def ShallowCopyDynamicPropertiesTo(self, target: Any, over_write: bool | None=None) -> None:
        this: DynamicObj = self
        over_write_1: bool = default_arg(over_write, False)
        def action(kv: Any) -> None:
            match_value: PropertyHelper | None = target.TryGetPropertyHelper(kv[0])
            if match_value is None:
                target.SetProperty(kv[0], kv[1])

            else: 
                def _arrow57(__unit: None=None, kv: Any=kv) -> bool:
                    pi: PropertyHelper = match_value
                    return over_write_1

                if _arrow57():
                    pi_1: PropertyHelper = match_value
                    pi_1.SetValue(target, kv[1])



        iterate(action, this.GetProperties(False))

    def ShallowCopyDynamicProperties(self, __unit: None=None) -> DynamicObj:
        this: DynamicObj = self
        target: DynamicObj = DynamicObj()
        this.ShallowCopyDynamicPropertiesTo(target, True)
        return target

    def DeepCopyPropertiesTo(self, target: Any, over_write: bool | None=None, include_instance_properties: bool | None=None) -> None:
        this: DynamicObj = self
        over_write_1: bool = default_arg(over_write, False)
        include_instance_properties_1: bool = default_arg(include_instance_properties, True)
        def action(kv: Any) -> None:
            match_value: PropertyHelper | None = target.TryGetPropertyHelper(kv[0])
            if match_value is None:
                target.SetProperty(kv[0], CopyUtils_tryDeepCopyObj_75B3D832(kv[1]))

            else: 
                def _arrow58(__unit: None=None, kv: Any=kv) -> bool:
                    pi: PropertyHelper = match_value
                    return over_write_1

                if _arrow58():
                    pi_1: PropertyHelper = match_value
                    pi_1.SetValue(target, CopyUtils_tryDeepCopyObj_75B3D832(kv[1]))



        iterate(action, this.GetProperties(include_instance_properties_1))

    def DeepCopyProperties(self, include_instance_properties: bool | None=None) -> Any:
        this: DynamicObj = self
        return CopyUtils_tryDeepCopyObj_75B3D832(this, default_arg(include_instance_properties, True))

    @staticmethod
    def op_dynamic(lookup: Any, name: str) -> Any:
        match_value: Any | None = lookup.TryGetPropertyValue(name)
        if match_value is None:
            raise Exception()

        else: 
            return value_2(match_value)


    @staticmethod
    def op_dynamic_assignment(lookup: Any, name: str, value: Any) -> None:
        lookup.SetProperty(name, value)

    def ReferenceEquals(self, other: DynamicObj) -> bool:
        this: DynamicObj = self
        return this is other

    def StructurallyEquals(self, other: DynamicObj) -> bool:
        this: DynamicObj = self
        return safe_hash(this) == safe_hash(other)

    def __hash__(self, __unit: None=None) -> int:
        this: DynamicObj = self
        return HashUtils_deepHash_4E60E31B(this)

    def __eq__(self, o: Any=None) -> bool:
        this: DynamicObj = self
        return this.StructurallyEquals(o) if isinstance(o, DynamicObj) else False


DynamicObj_reflection = _expr59

def DynamicObj__ctor(__unit: None=None) -> DynamicObj:
    return DynamicObj(__unit)


def _expr60() -> TypeInfo:
    return class_type("DynamicObj.HashUtils", None, HashUtils)


class HashUtils:
    ...

HashUtils_reflection = _expr60

def _expr61() -> TypeInfo:
    return class_type("DynamicObj.CopyUtils", None, CopyUtils)


class CopyUtils:
    ...

CopyUtils_reflection = _expr61

def HashUtils_deepHash_4E60E31B(o: Any=None) -> int:
    if isinstance(o, DynamicObj):
        def f(o_2: Any=None, o: Any=o) -> int:
            return HashUtils_deepHash_4E60E31B(o_2)

        def projection(pair: Any, o: Any=o) -> str:
            return pair[0]

        class ObjectExpr62:
            @property
            def Compare(self) -> Callable[[str, str], int]:
                return compare_primitives

        return box_hash_key_val_seq_by(f, sort_by(projection, o.GetProperties(True), ObjectExpr62()))

    elif str(type(o)) == "<class \'str\'>":
        return hash_1_1(o)

    elif is_iterable(o):
        en: IEnumerator[Any] = get_enumerator(o)
        def _arrow65(__unit: None=None, o: Any=o) -> IEnumerable_1[int]:
            def _arrow63(__unit: None=None) -> bool:
                return en.System_Collections_IEnumerator_MoveNext()

            def _arrow64(__unit: None=None) -> IEnumerable_1[int]:
                return singleton(HashUtils_deepHash_4E60E31B(en.System_Collections_IEnumerator_get_Current()))

            return enumerate_while(_arrow63, delay(_arrow64))

        l: FSharpList[int] = to_list(delay(_arrow65))
        if is_empty(l):
            return 0

        else: 
            def _arrow66(hash_1: int, hash_2: int, o: Any=o) -> int:
                return merge_hashes(hash_1, hash_2)

            return reduce(_arrow66, l)


    else: 
        return hash_1_1(o)



def CopyUtils_tryDeepCopyObj_75B3D832(o: Any=None, include_instance_properties: bool | None=None) -> Any:
    include_instance_properties_1: bool = default_arg(include_instance_properties, True)
    def try_deep_copy_obj(o_1: Any=None, o: Any=o, include_instance_properties: Any=include_instance_properties) -> Any:
        (pattern_matching_result, o_7) = (None, None)
        if str(type(o_1)) == "<class \'bool\'>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<fable_modules.fable_library.types.uint8\'>>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'fable_modules.fable_library.types.int8\'>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'fable_modules.fable_library.types.int16\'>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'fable_modules.fable_library.types.uint16\'>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'int\'>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'fable_modules.fable_library.types.uint32>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'fable_modules.fable_library.types.int64\'>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'fable_modules.fable_library.types.uint32\'>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'float\'>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'fable_modules.fable_library.types.float32\'>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'str\'>":
            pattern_matching_result = 0

        elif str(type(o_1)) == "<class \'str\'>":
            pattern_matching_result = 0

        elif o_1 is None:
            pattern_matching_result = 0

        elif isinstance(o_1, dict):
            pattern_matching_result = 1

        elif is_array_like(o_1):
            pattern_matching_result = 2

        elif isinstance(o_1, FSharpList):
            pattern_matching_result = 3

        elif isinstance(o_1, DynamicObj):
            if hasattr(o_1, 'System_ICloneable_Clone') and callable(o_1.System_ICloneable_Clone):
                pattern_matching_result = 4
                o_7 = o_1

            else: 
                pattern_matching_result = 5


        elif hasattr(o_1, 'System_ICloneable_Clone') and callable(o_1.System_ICloneable_Clone):
            pattern_matching_result = 4
            o_7 = o_1

        else: 
            pattern_matching_result = 6

        if pattern_matching_result == 0:
            return o_1

        elif pattern_matching_result == 1:
            class ObjectExpr67:
                @property
                def Equals(self) -> Callable[[Any, Any], bool]:
                    return equals

                @property
                def GetHashCode(self) -> Callable[[Any], int]:
                    return structural_hash

            new_dict: Any = Dictionary([], ObjectExpr67())
            enumerator: Any = get_enumerator(o_1)
            try: 
                while enumerator.System_Collections_IEnumerator_MoveNext():
                    kv: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                    add_to_dict(new_dict, try_deep_copy_obj(kv[0]), try_deep_copy_obj(kv[1]))

            finally: 
                dispose(enumerator)

            return new_dict

        elif pattern_matching_result == 2:
            def _arrow68(__unit: None=None, o_1: Any=o_1) -> IEnumerable_1[DynamicObj]:
                return map_1(try_deep_copy_obj, o_1)

            return list(to_list(delay(_arrow68)))

        elif pattern_matching_result == 3:
            def _arrow69(__unit: None=None, o_1: Any=o_1) -> IEnumerable_1[DynamicObj]:
                return map_1(try_deep_copy_obj, o_1)

            return to_list(delay(_arrow69))

        elif pattern_matching_result == 4:
            return o_7.System_ICloneable_Clone()

        elif pattern_matching_result == 5:
            new_dyn: DynamicObj = DynamicObj()
            with get_enumerator(o_1.GetProperties(include_instance_properties_1)) as enumerator_1:
                while enumerator_1.System_Collections_IEnumerator_MoveNext():
                    kv_1: Any = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current()
                    new_dyn.SetProperty(kv_1[0], try_deep_copy_obj(kv_1[1]))
            return new_dyn

        elif pattern_matching_result == 6:
            return o_1


    return try_deep_copy_obj(o)


__all__ = ["DynamicObj_reflection", "HashUtils_reflection", "CopyUtils_reflection", "HashUtils_deepHash_4E60E31B", "CopyUtils_tryDeepCopyObj_75B3D832"]




from __future__ import annotations

from .types import int64


def to_fixed(x: float, dp: int | None = None) -> str:
    x = int(x) if int(x) == x else x

    if dp is not None:
        fmt = "{:.%sf}" % dp
        return fmt.format(x)

    return f"{x}"


def to_precision(x: float, sd: int | None = None) -> str:
    x = int(x) if int(x) == x else x

    if sd is not None:
        fmt = "{:.%se}" % sd
        return fmt.format(x)

    return f"{x}"


def to_exponential(x: float, dp: int | None = None) -> str:
    if dp is not None:
        fmt = "{:.%se}" % dp
        return fmt.format(x)

    return f"{x}"


def to_hex(x: int) -> str:
    def rshift(val: int, n: int) -> int:
        sign = 0x10000000000000000 if isinstance(val, int64) else 0x100000000
        return val >> n if val >= 0 else (val + sign) >> n

    return f"{rshift(x, 0):x}"


def multiply(x: float, y: float) -> float:
    return x * y


__all__ = ["to_fixed", "to_precision", "to_exponential", "to_hex", "multiply"]



from __future__ import annotations
from abc import abstractmethod
from typing import (Protocol, TypeVar)
from .futures import Future_1

_T = TypeVar("_T")

class AbstractEventLoop(Protocol):
    @abstractmethod
    def create_future(self) -> Future_1[_T]:
        ...

    @abstractmethod
    def run_forever(self) -> None:
        ...





from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..arctrl_contract.contract import (Contract, DTOType, DTO)
from ..arctrl_core.Helper.hash_codes import box_hash_array
from ..arctrl_file_system.path import alternative_licensefile_names
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (contains, for_all)
from ..fable_library.types import Array
from ..fable_library.util import (string_hash, to_enumerable)

def _expr3768() -> TypeInfo:
    return class_type("ARCtrl.License", None, License)


class License:
    def __init__(self, content_type: str, content: str, path: str | None=None) -> None:
        self._type: str = content_type
        self._content: str = content
        self._staticHash: int = 0
        self._path: str = "LICENSE" if (path is None) else path

    @property
    def Type(self, __unit: None=None) -> str:
        this: License = self
        return this._type

    @Type.setter
    def Type(self, h: str) -> None:
        this: License = self
        this._type = h

    @property
    def Content(self, __unit: None=None) -> str:
        this: License = self
        return this._content

    @Content.setter
    def Content(self, h: str) -> None:
        this: License = self
        this._content = h

    @property
    def StaticHash(self, __unit: None=None) -> int:
        this: License = self
        return this._staticHash

    @StaticHash.setter
    def StaticHash(self, h: int) -> None:
        this: License = self
        this._staticHash = h or 0

    @property
    def Path(self, __unit: None=None) -> str:
        this: License = self
        return this._path

    @Path.setter
    def Path(self, p: str) -> None:
        this: License = self
        this._path = p

    @staticmethod
    def init_fulltext(content: str, path: str | None=None) -> License:
        return License("fulltext", content, path)

    def ToCreateContract(self, __unit: None=None) -> Contract:
        this: License = self
        match_value: str = this.Type
        return Contract.create_create(this._path, DTOType(10), DTO(1, this.Content))

    def ToUpdateContract(self, __unit: None=None) -> Contract:
        this: License = self
        match_value: str = this.Type
        return Contract.create_update(this._path, DTOType(10), DTO(1, this.Content))

    def ToDeleteContract(self, __unit: None=None) -> Contract:
        this: License = self
        return Contract.create_delete(this._path)

    def GetRenameContracts(self, new_path: str) -> Array[Contract]:
        this: License = self
        delete_contract: Contract = this.ToDeleteContract()
        this.Path = new_path
        def _arrow3762(__unit: None=None) -> Contract:
            match_value: str = this.Type
            return Contract.create_create(new_path, DTOType(10), DTO(1, this.Content))

        return [delete_contract, _arrow3762()]

    @staticmethod
    def to_delete_contract(license: License) -> Contract:
        return license.ToDeleteContract()

    @staticmethod
    def to_create_contract(license: License) -> Contract:
        return license.ToCreateContract()

    @staticmethod
    def to_update_contract(license: License) -> Contract:
        return license.ToUpdateContract()

    @staticmethod
    def get_rename_contracts(new_path: str) -> Callable[[License], Array[Contract]]:
        def _arrow3763(license: License) -> Array[Contract]:
            return license.GetRenameContracts(new_path)

        return _arrow3763

    @staticmethod
    def try_from_read_contract(c: Contract) -> License | None:
        def _arrow3766(__unit: None=None) -> bool:
            txt: str = c.DTO.fields[0]
            class ObjectExpr3765:
                @property
                def Equals(self) -> Callable[[str, str], bool]:
                    def _arrow3764(x: str, y: str) -> bool:
                        return x == y

                    return _arrow3764

                @property
                def GetHashCode(self) -> Callable[[str], int]:
                    return string_hash

            return True if (c.Path == "LICENSE") else contains(c.Path, alternative_licensefile_names, ObjectExpr3765())

        def _arrow3767(__unit: None=None) -> License | None:
            txt_1: str = c.DTO.fields[0]
            return License.init_fulltext(txt_1, c.Path)

        return (((((_arrow3767() if _arrow3766() else None) if (c.DTO.tag == 1) else None) if (c.DTO is not None) else None) if (c.DTOType.tag == 10) else None) if (c.DTOType is not None) else None) if (c.Operation == "READ") else None

    @staticmethod
    def GetDefaultLicense(__unit: None=None) -> License:
        return License.init_fulltext("ALL RIGHTS RESERVED BY THE AUTHORS")

    def Copy(self, __unit: None=None) -> License:
        this: License = self
        return License(this.Type, this.Content)

    def __eq__(self, other: Any=None) -> bool:
        this: License = self
        return this.StructurallyEquals(other) if isinstance(other, License) else False

    def __hash__(self, __unit: None=None) -> Any:
        this: License = self
        return box_hash_array([this.Type, this.Content, this.Path])

    def StructurallyEquals(self, other: License) -> bool:
        this: License = self
        def predicate(x: bool) -> bool:
            return x == True

        return for_all(predicate, to_enumerable([this.Type == other.Type, this.Content == other.Content, this.Path == other.Path]))

    def ReferenceEquals(self, other: License) -> bool:
        this: License = self
        return this is other


License_reflection = _expr3768

def License__ctor_Z2FC25A28(content_type: str, content: str, path: str | None=None) -> License:
    return License(content_type, content, path)


__all__ = ["License_reflection"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ...arctrl_core.arc_types import (ArcStudy, ArcAssay)
from ...arctrl_core.comment import (Comment, Comment_reflection, Remark)
from ...arctrl_core.conversion import (ARCtrl_ArcTable__ArcTable_GetProtocols, ARCtrl_ArcTable__ArcTable_GetProcesses)
from ...arctrl_core.Helper.collections_ import (ResizeArray_iter, Option_fromValueWithDefault)
from ...arctrl_core.Helper.identifier import (create_missing_identifier, Study_fileNameFromIdentifier)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.person import Person
from ...arctrl_core.Process.factor import Factor
from ...arctrl_core.Process.process_sequence import get_factors
from ...arctrl_core.Process.protocol import Protocol
from ...arctrl_core.publication import Publication
from ...arctrl_core.Table.arc_table import ArcTable
from ...fable_library.list import (FSharpList, of_array, map, empty, cons, reverse, append, of_seq)
from ...fable_library.map_util import add_to_dict
from ...fable_library.option import default_arg
from ...fable_library.reflection import (TypeInfo, string_type, list_type, record_type)
from ...fable_library.seq import (collect, delay, append as append_1, singleton)
from ...fable_library.seq2 import List_distinct
from ...fable_library.string_ import starts_with_exact
from ...fable_library.types import (Record, Array)
from ...fable_library.util import (string_hash, IEnumerable_1, IEnumerator, to_enumerable)
from .assays import (from_rows as from_rows_4, to_rows as to_rows_4)
from .comment import (Comment_fromString, Comment_toString)
from .contacts import (from_rows as from_rows_6, to_rows as to_rows_6)
from .design_descriptors import (from_rows as from_rows_1, to_rows as to_rows_1)
from .factors import (from_rows as from_rows_3, to_rows as to_rows_3)
from .protocols import (from_rows as from_rows_5, to_rows as to_rows_5)
from .publication import (from_rows as from_rows_2, to_rows as to_rows_2)
from .sparse_table import (SparseTable__TryGetValueDefault_5BAE6133, SparseTable, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1, SparseRowModule_fromValues)

def _expr3468() -> TypeInfo:
    return record_type("ARCtrl.Spreadsheet.Studies.StudyInfo", [], StudyInfo, lambda: [("Identifier", string_type), ("Title", string_type), ("Description", string_type), ("SubmissionDate", string_type), ("PublicReleaseDate", string_type), ("FileName", string_type), ("Comments", list_type(Comment_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class StudyInfo(Record):
    Identifier: str
    Title: str
    Description: str
    SubmissionDate: str
    PublicReleaseDate: str
    FileName: str
    Comments: FSharpList[Comment]

StudyInfo_reflection = _expr3468

def StudyInfo_create(identifier: str, title: str, description: str, submission_date: str, public_release_date: str, file_name: str, comments: FSharpList[Comment]) -> StudyInfo:
    return StudyInfo(identifier, title, description, submission_date, public_release_date, file_name, comments)


def StudyInfo_get_Labels(__unit: None=None) -> FSharpList[str]:
    return of_array(["Study Identifier", "Study Title", "Study Description", "Study Submission Date", "Study Public Release Date", "Study File Name"])


def StudyInfo_FromSparseTable_3ECCA699(matrix: SparseTable) -> StudyInfo:
    def mapping(k: str, matrix: Any=matrix) -> Comment:
        return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, 0)))

    comments: FSharpList[Comment] = map(mapping, matrix.CommentKeys)
    return StudyInfo_create(SparseTable__TryGetValueDefault_5BAE6133(matrix, create_missing_identifier(), ("Study Identifier", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Study Title", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Study Description", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Study Submission Date", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Study Public Release Date", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Study File Name", 0)), comments)


def StudyInfo_ToSparseTable_1680536E(study: ArcStudy) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, StudyInfo_get_Labels(), None, 2)
    comment_keys: FSharpList[str] = empty()
    pattern_input: tuple[str, str] = (("", "")) if starts_with_exact(study.Identifier, "MISSING_IDENTIFIER_") else ((study.Identifier, Study_fileNameFromIdentifier(study.Identifier)))
    add_to_dict(matrix.Matrix, ("Study Identifier", 1), pattern_input[0])
    add_to_dict(matrix.Matrix, ("Study Title", 1), default_arg(study.Title, ""))
    add_to_dict(matrix.Matrix, ("Study Description", 1), default_arg(study.Description, ""))
    add_to_dict(matrix.Matrix, ("Study Submission Date", 1), default_arg(study.SubmissionDate, ""))
    add_to_dict(matrix.Matrix, ("Study Public Release Date", 1), default_arg(study.PublicReleaseDate, ""))
    add_to_dict(matrix.Matrix, ("Study File Name", 1), pattern_input[1])
    def f(comment: Comment, study: Any=study) -> None:
        nonlocal comment_keys
        pattern_input_1: tuple[str, str] = Comment_toString(comment)
        n: str = pattern_input_1[0]
        comment_keys = cons(n, comment_keys)
        add_to_dict(matrix.Matrix, (n, 1), pattern_input_1[1])

    ResizeArray_iter(f, study.Comments)
    class ObjectExpr3470:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3469(x: str, y: str) -> bool:
                return x == y

            return _arrow3469

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3470())), matrix.ColumnCount)


def StudyInfo_fromRows(line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], StudyInfo]:
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, StudyInfo_get_Labels(), line_number)
    return (tupled_arg[0], tupled_arg[1], tupled_arg[2], StudyInfo_FromSparseTable_3ECCA699(tupled_arg[3]))


def StudyInfo_toRows_1680536E(study: ArcStudy) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    return SparseTable_ToRows_759CAFC1(StudyInfo_ToSparseTable_1680536E(study))


def from_parts(study_info: StudyInfo, design_descriptors: FSharpList[OntologyAnnotation], publications: FSharpList[Publication], factors: FSharpList[Factor], assays: FSharpList[ArcAssay], protocols: FSharpList[Protocol], contacts: FSharpList[Person]) -> tuple[ArcStudy, FSharpList[ArcAssay]] | None:
    def mapping(assay: ArcAssay, study_info: Any=study_info, design_descriptors: Any=design_descriptors, publications: Any=publications, factors: Any=factors, assays: Any=assays, protocols: Any=protocols, contacts: Any=contacts) -> str:
        return assay.Identifier

    assay_identifiers: FSharpList[str] = map(mapping, assays)
    arcstudy: ArcStudy
    title: str | None = Option_fromValueWithDefault("", study_info.Title)
    description: str | None = Option_fromValueWithDefault("", study_info.Description)
    submission_date: str | None = Option_fromValueWithDefault("", study_info.SubmissionDate)
    public_release_date: str | None = Option_fromValueWithDefault("", study_info.PublicReleaseDate)
    publications_1: Array[Publication] = list(publications)
    contacts_1: Array[Person] = list(contacts)
    study_design_descriptors: Array[OntologyAnnotation] = list(design_descriptors)
    registered_assay_identifiers: Array[str] = list(assay_identifiers)
    comments: Array[Comment] = list(study_info.Comments)
    arcstudy = ArcStudy.make(study_info.Identifier, title, description, submission_date, public_release_date, publications_1, contacts_1, study_design_descriptors, [], None, registered_assay_identifiers, comments)
    if (arcstudy.Identifier == "") if arcstudy.is_empty else False:
        return None

    else: 
        return (arcstudy, assays)



def from_rows(line_number: int, en: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], tuple[ArcStudy, FSharpList[ArcAssay]] | None]:
    def loop(last_line_mut: str | None, study_info_mut: StudyInfo, design_descriptors_mut: FSharpList[OntologyAnnotation], publications_mut: FSharpList[Publication], factors_mut: FSharpList[Factor], assays_mut: FSharpList[ArcAssay], protocols_mut: FSharpList[Protocol], contacts_mut: FSharpList[Person], remarks_mut: FSharpList[Remark], line_number_1_mut: int, line_number: Any=line_number, en: Any=en) -> tuple[str | None, int, FSharpList[Remark], tuple[ArcStudy, FSharpList[ArcAssay]] | None]:
        while True:
            (last_line, study_info, design_descriptors, publications, factors, assays, protocols, contacts, remarks, line_number_1) = (last_line_mut, study_info_mut, design_descriptors_mut, publications_mut, factors_mut, assays_mut, protocols_mut, contacts_mut, remarks_mut, line_number_1_mut)
            (pattern_matching_result, k_12) = (None, None)
            if last_line is not None:
                if last_line == "STUDY DESIGN DESCRIPTORS":
                    pattern_matching_result = 0

                elif last_line == "STUDY PUBLICATIONS":
                    pattern_matching_result = 1

                elif last_line == "STUDY FACTORS":
                    pattern_matching_result = 2

                elif last_line == "STUDY ASSAYS":
                    pattern_matching_result = 3

                elif last_line == "STUDY PROTOCOLS":
                    pattern_matching_result = 4

                elif last_line == "STUDY CONTACTS":
                    pattern_matching_result = 5

                else: 
                    pattern_matching_result = 6
                    k_12 = last_line


            else: 
                pattern_matching_result = 6
                k_12 = last_line

            if pattern_matching_result == 0:
                pattern_input: tuple[str | None, int, FSharpList[Remark], FSharpList[OntologyAnnotation]] = from_rows_1("Study Design", line_number_1 + 1, en)
                last_line_mut = pattern_input[0]
                study_info_mut = study_info
                design_descriptors_mut = pattern_input[3]
                publications_mut = publications
                factors_mut = factors
                assays_mut = assays
                protocols_mut = protocols
                contacts_mut = contacts
                remarks_mut = append(remarks, pattern_input[2])
                line_number_1_mut = pattern_input[1]
                continue

            elif pattern_matching_result == 1:
                pattern_input_1: tuple[str | None, int, FSharpList[Remark], FSharpList[Publication]] = from_rows_2("Study Publication", line_number_1 + 1, en)
                last_line_mut = pattern_input_1[0]
                study_info_mut = study_info
                design_descriptors_mut = design_descriptors
                publications_mut = pattern_input_1[3]
                factors_mut = factors
                assays_mut = assays
                protocols_mut = protocols
                contacts_mut = contacts
                remarks_mut = append(remarks, pattern_input_1[2])
                line_number_1_mut = pattern_input_1[1]
                continue

            elif pattern_matching_result == 2:
                pattern_input_2: tuple[str | None, int, FSharpList[Remark], FSharpList[Factor]] = from_rows_3("Study Factor", line_number_1 + 1, en)
                last_line_mut = pattern_input_2[0]
                study_info_mut = study_info
                design_descriptors_mut = design_descriptors
                publications_mut = publications
                factors_mut = pattern_input_2[3]
                assays_mut = assays
                protocols_mut = protocols
                contacts_mut = contacts
                remarks_mut = append(remarks, pattern_input_2[2])
                line_number_1_mut = pattern_input_2[1]
                continue

            elif pattern_matching_result == 3:
                pattern_input_3: tuple[str | None, int, FSharpList[Remark], FSharpList[ArcAssay]] = from_rows_4("Study Assay", line_number_1 + 1, en)
                last_line_mut = pattern_input_3[0]
                study_info_mut = study_info
                design_descriptors_mut = design_descriptors
                publications_mut = publications
                factors_mut = factors
                assays_mut = pattern_input_3[3]
                protocols_mut = protocols
                contacts_mut = contacts
                remarks_mut = append(remarks, pattern_input_3[2])
                line_number_1_mut = pattern_input_3[1]
                continue

            elif pattern_matching_result == 4:
                pattern_input_4: tuple[str | None, int, FSharpList[Remark], FSharpList[Protocol]] = from_rows_5("Study Protocol", line_number_1 + 1, en)
                last_line_mut = pattern_input_4[0]
                study_info_mut = study_info
                design_descriptors_mut = design_descriptors
                publications_mut = publications
                factors_mut = factors
                assays_mut = assays
                protocols_mut = pattern_input_4[3]
                contacts_mut = contacts
                remarks_mut = append(remarks, pattern_input_4[2])
                line_number_1_mut = pattern_input_4[1]
                continue

            elif pattern_matching_result == 5:
                pattern_input_5: tuple[str | None, int, FSharpList[Remark], FSharpList[Person]] = from_rows_6("Study Person", line_number_1 + 1, en)
                last_line_mut = pattern_input_5[0]
                study_info_mut = study_info
                design_descriptors_mut = design_descriptors
                publications_mut = publications
                factors_mut = factors
                assays_mut = assays
                protocols_mut = protocols
                contacts_mut = pattern_input_5[3]
                remarks_mut = append(remarks, pattern_input_5[2])
                line_number_1_mut = pattern_input_5[1]
                continue

            elif pattern_matching_result == 6:
                return (k_12, line_number_1, remarks, from_parts(study_info, design_descriptors, publications, factors, assays, protocols, contacts))

            break

    pattern_input_6: tuple[str | None, int, FSharpList[Remark], StudyInfo] = StudyInfo_fromRows(line_number, en)
    return loop(pattern_input_6[0], pattern_input_6[3], empty(), empty(), empty(), empty(), empty(), empty(), pattern_input_6[2], pattern_input_6[1])


def to_rows(study: ArcStudy, assays: FSharpList[ArcAssay] | None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    def mapping(p: ArcTable, study: Any=study, assays: Any=assays) -> FSharpList[Protocol]:
        return ARCtrl_ArcTable__ArcTable_GetProtocols(p)

    protocols: FSharpList[Protocol] = of_seq(collect(mapping, study.Tables))
    def mapping_1(f: ArcTable, study: Any=study, assays: Any=assays) -> FSharpList[Factor]:
        return get_factors(ARCtrl_ArcTable__ArcTable_GetProcesses(f))

    factors: FSharpList[Factor] = of_seq(collect(mapping_1, study.Tables))
    assays_1: FSharpList[ArcAssay] = default_arg(assays, of_seq(study.GetRegisteredAssaysOrIdentifier()))
    def _arrow3483(__unit: None=None, study: Any=study, assays: Any=assays) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
        def _arrow3482(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
            def _arrow3481(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                def _arrow3480(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                    def _arrow3479(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                        def _arrow3478(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                            def _arrow3477(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                def _arrow3476(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                    def _arrow3475(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                        def _arrow3474(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                            def _arrow3473(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                                def _arrow3472(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                                    def _arrow3471(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                                        return to_rows_6("Study Person", of_seq(study.Contacts))

                                                    return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["STUDY CONTACTS"]))), delay(_arrow3471))

                                                return append_1(to_rows_5("Study Protocol", protocols), delay(_arrow3472))

                                            return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["STUDY PROTOCOLS"]))), delay(_arrow3473))

                                        return append_1(to_rows_4("Study Assay", assays_1), delay(_arrow3474))

                                    return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["STUDY ASSAYS"]))), delay(_arrow3475))

                                return append_1(to_rows_3("Study Factor", factors), delay(_arrow3476))

                            return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["STUDY FACTORS"]))), delay(_arrow3477))

                        return append_1(to_rows_2("Study Publication", of_seq(study.Publications)), delay(_arrow3478))

                    return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["STUDY PUBLICATIONS"]))), delay(_arrow3479))

                return append_1(to_rows_1("Study Design", of_seq(study.StudyDesignDescriptors)), delay(_arrow3480))

            return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["STUDY DESIGN DESCRIPTORS"]))), delay(_arrow3481))

        return append_1(StudyInfo_toRows_1680536E(study), delay(_arrow3482))

    return delay(_arrow3483)


__all__ = ["StudyInfo_reflection", "StudyInfo_create", "StudyInfo_get_Labels", "StudyInfo_FromSparseTable_3ECCA699", "StudyInfo_ToSparseTable_1680536E", "StudyInfo_fromRows", "StudyInfo_toRows_1680536E", "from_parts", "from_rows", "to_rows"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ..arctrl_core.arc_types import ArcRun
from ..arctrl_core.comment import Comment
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.person import Person
from ..arctrl_core.Table.arc_table import ArcTable
from ..arctrl_core.Table.composite_cell import CompositeCell
from ..fable_library.list import (choose, of_array, FSharpList)
from ..fable_library.option import map
from ..fable_library.seq import map as map_1
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.decode import (object, IRequiredGetter, string, IOptionalGetter, resize_array, IGetters)
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .comment import (encoder as encoder_4, decoder as decoder_4)
from .DataMap.data_map import (encoder as encoder_1, decoder as decoder_2, encoder_compressed as encoder_compressed_1, decoder_compressed as decoder_compressed_2)
from .encode import (try_include, try_include_seq)
from .ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder)
from .person import (encoder as encoder_3, decoder as decoder_3)
from .Table.arc_table import (encoder as encoder_2, decoder as decoder_1, encoder_compressed as encoder_compressed_2, decoder_compressed as decoder_compressed_1)

__A_ = TypeVar("__A_")

def encoder(run: ArcRun) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], run: Any=run) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3096(__unit: None=None, run: Any=run) -> IEncodable:
        value: str = run.Identifier
        class ObjectExpr3095(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3095()

    def _arrow3098(value_1: str, run: Any=run) -> IEncodable:
        class ObjectExpr3097(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr3097()

    def _arrow3100(value_3: str, run: Any=run) -> IEncodable:
        class ObjectExpr3099(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_3)

        return ObjectExpr3099()

    def _arrow3103(oa: OntologyAnnotation, run: Any=run) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def _arrow3104(oa_1: OntologyAnnotation, run: Any=run) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow3105(oa_2: OntologyAnnotation, run: Any=run) -> IEncodable:
        return OntologyAnnotation_encoder(oa_2)

    def _arrow3108(dm: DataMap, run: Any=run) -> IEncodable:
        return encoder_1(dm)

    def _arrow3112(value_5: str, run: Any=run) -> IEncodable:
        class ObjectExpr3111(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr3111()

    def _arrow3113(table: ArcTable, run: Any=run) -> IEncodable:
        return encoder_2(table)

    def _arrow3114(person: Person, run: Any=run) -> IEncodable:
        return encoder_3(person)

    def _arrow3115(comment: Comment, run: Any=run) -> IEncodable:
        return encoder_4(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("Identifier", _arrow3096()), try_include("Title", _arrow3098, run.Title), try_include("Description", _arrow3100, run.Description), try_include("MeasurementType", _arrow3103, run.MeasurementType), try_include("TechnologyType", _arrow3104, run.TechnologyType), try_include("TechnologyPlatform", _arrow3105, run.TechnologyPlatform), try_include("DataMap", _arrow3108, run.DataMap), try_include_seq("WorkflowIdentifiers", _arrow3112, run.WorkflowIdentifiers), try_include_seq("Tables", _arrow3113, run.Tables), try_include_seq("Performers", _arrow3114, run.Performers), try_include_seq("Comments", _arrow3115, run.Comments)]))
    class ObjectExpr3120(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], run: Any=run) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr3120()


def _arrow3143(get: IGetters) -> ArcRun:
    def _arrow3130(__unit: None=None) -> str:
        object_arg: IRequiredGetter = get.Required
        return object_arg.Field("Identifier", string)

    def _arrow3132(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("Title", string)

    def _arrow3133(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("Description", string)

    def _arrow3134(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("MeasurementType", OntologyAnnotation_decoder)

    def _arrow3135(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("TechnologyType", OntologyAnnotation_decoder)

    def _arrow3136(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("TechnologyPlatform", OntologyAnnotation_decoder)

    def _arrow3137(__unit: None=None) -> Array[str] | None:
        arg_13: Decoder_1[Array[str]] = resize_array(string)
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("WorkflowIdentifiers", arg_13)

    def _arrow3139(__unit: None=None) -> Array[ArcTable] | None:
        arg_15: Decoder_1[Array[ArcTable]] = resize_array(decoder_1)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("Tables", arg_15)

    def _arrow3140(__unit: None=None) -> DataMap | None:
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("DataMap", decoder_2)

    def _arrow3141(__unit: None=None) -> Array[Person] | None:
        arg_19: Decoder_1[Array[Person]] = resize_array(decoder_3)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("Performers", arg_19)

    def _arrow3142(__unit: None=None) -> Array[Comment] | None:
        arg_21: Decoder_1[Array[Comment]] = resize_array(decoder_4)
        object_arg_10: IOptionalGetter = get.Optional
        return object_arg_10.Field("Comments", arg_21)

    return ArcRun.create(_arrow3130(), _arrow3132(), _arrow3133(), _arrow3134(), _arrow3135(), _arrow3136(), _arrow3137(), _arrow3139(), _arrow3140(), _arrow3141(), None, None, _arrow3142())


decoder: Decoder_1[ArcRun] = object(_arrow3143)

def encoder_compressed(string_table: Any, oa_table: Any, cell_table: Any, run: ArcRun) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3151(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        value: str = run.Identifier
        class ObjectExpr3150(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3150()

    def _arrow3155(value_1: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        class ObjectExpr3153(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr3153()

    def _arrow3158(value_3: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        class ObjectExpr3157(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_3)

        return ObjectExpr3157()

    def _arrow3160(oa: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def _arrow3161(oa_1: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow3162(oa_2: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        return OntologyAnnotation_encoder(oa_2)

    def _arrow3163(dm: DataMap, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        return encoder_compressed_1(string_table, oa_table, cell_table, dm)

    def _arrow3167(value_5: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        class ObjectExpr3166(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr3166()

    def _arrow3169(table: ArcTable, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        return encoder_compressed_2(string_table, oa_table, cell_table, table)

    def _arrow3170(person: Person, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        return encoder_3(person)

    def _arrow3172(comment: Comment, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> IEncodable:
        return encoder_4(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("Identifier", _arrow3151()), try_include("Title", _arrow3155, run.Title), try_include("Description", _arrow3158, run.Description), try_include("MeasurementType", _arrow3160, run.MeasurementType), try_include("TechnologyType", _arrow3161, run.TechnologyType), try_include("TechnologyPlatform", _arrow3162, run.TechnologyPlatform), try_include("DataMap", _arrow3163, run.DataMap), try_include_seq("WorkflowIdentifiers", _arrow3167, run.WorkflowIdentifiers), try_include_seq("Tables", _arrow3169, run.Tables), try_include_seq("Performers", _arrow3170, run.Performers), try_include_seq("Comments", _arrow3172, run.Comments)]))
    class ObjectExpr3174(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, run: Any=run) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr3174()


def decoder_compressed(string_table: Array[str], oa_table: Array[OntologyAnnotation], cell_table: Array[CompositeCell]) -> Decoder_1[ArcRun]:
    def _arrow3191(get: IGetters, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> ArcRun:
        def _arrow3175(__unit: None=None) -> str:
            object_arg: IRequiredGetter = get.Required
            return object_arg.Field("Identifier", string)

        def _arrow3176(__unit: None=None) -> str | None:
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("Title", string)

        def _arrow3177(__unit: None=None) -> str | None:
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("Description", string)

        def _arrow3180(__unit: None=None) -> OntologyAnnotation | None:
            object_arg_3: IOptionalGetter = get.Optional
            return object_arg_3.Field("MeasurementType", OntologyAnnotation_decoder)

        def _arrow3181(__unit: None=None) -> OntologyAnnotation | None:
            object_arg_4: IOptionalGetter = get.Optional
            return object_arg_4.Field("TechnologyType", OntologyAnnotation_decoder)

        def _arrow3182(__unit: None=None) -> OntologyAnnotation | None:
            object_arg_5: IOptionalGetter = get.Optional
            return object_arg_5.Field("TechnologyPlatform", OntologyAnnotation_decoder)

        def _arrow3183(__unit: None=None) -> Array[str] | None:
            arg_13: Decoder_1[Array[str]] = resize_array(string)
            object_arg_6: IOptionalGetter = get.Optional
            return object_arg_6.Field("WorkflowIdentifiers", arg_13)

        def _arrow3184(__unit: None=None) -> Array[ArcTable] | None:
            arg_15: Decoder_1[Array[ArcTable]] = resize_array(decoder_compressed_1(string_table, oa_table, cell_table))
            object_arg_7: IOptionalGetter = get.Optional
            return object_arg_7.Field("Tables", arg_15)

        def _arrow3187(__unit: None=None) -> DataMap | None:
            arg_17: Decoder_1[DataMap] = decoder_compressed_2(string_table, oa_table, cell_table)
            object_arg_8: IOptionalGetter = get.Optional
            return object_arg_8.Field("DataMap", arg_17)

        def _arrow3188(__unit: None=None) -> Array[Person] | None:
            arg_19: Decoder_1[Array[Person]] = resize_array(decoder_3)
            object_arg_9: IOptionalGetter = get.Optional
            return object_arg_9.Field("Performers", arg_19)

        def _arrow3190(__unit: None=None) -> Array[Comment] | None:
            arg_21: Decoder_1[Array[Comment]] = resize_array(decoder_4)
            object_arg_10: IOptionalGetter = get.Optional
            return object_arg_10.Field("Comments", arg_21)

        return ArcRun.create(_arrow3175(), _arrow3176(), _arrow3177(), _arrow3180(), _arrow3181(), _arrow3182(), _arrow3183(), _arrow3184(), _arrow3187(), _arrow3188(), None, None, _arrow3190())

    return object(_arrow3191)


__all__ = ["encoder", "decoder", "encoder_compressed", "decoder_compressed"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.map_util import (add_to_set, add_to_dict, get_item_from_dict, remove_from_dict)
from ...fable_library.option import (some, value as value_2, default_arg)
from ...fable_library.range import range_big_int
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import (initialize, delay, collect, singleton, reduce, map, iterate_indexed, iterate, length, max, find, try_find, try_pick, choose, to_list, exists)
from ...fable_library.string_ import (to_text, printf, replace)
from ...fable_library.util import (equals, IEnumerable_1, int32_to_string, ignore, compare_primitives, get_enumerator)
from ..Cells.fs_cell import (FsCell, DataType)
from ..Cells.fs_cells_collection import (FsCellsCollection, Dictionary_tryGet)
from ..fs_address import FsAddress
from ..fs_column import FsColumn
from ..fs_row import FsRow
from ..Ranges.fs_range import FsRange
from ..Ranges.fs_range_address import FsRangeAddress
from ..Ranges.fs_range_base import (FsRangeBase, FsRangeBase_reflection)
from ..Ranges.fs_range_column import FsRangeColumn
from ..Ranges.fs_range_row import FsRangeRow
from .fs_table_field import FsTableField

def _expr250() -> TypeInfo:
    return class_type("FsSpreadsheet.FsTable", None, FsTable, FsRangeBase_reflection())


class FsTable(FsRangeBase):
    def __init__(self, name: str, range_address: FsRangeAddress, show_totals_row: bool | None=None, show_header_row: bool | None=None, field_names: IEnumerable_1[str] | None=None) -> None:
        super().__init__(range_address)
        self._name: str = replace(name.strip(), " ", "_")
        self._lastRangeAddress: FsRangeAddress = range_address
        self._showTotalsRow: bool = default_arg(show_totals_row, False)
        self._showHeaderRow: bool = default_arg(show_header_row, True)
        def _arrow249(__unit: None=None) -> Any:
            fns: IEnumerable_1[str] = field_names
            if length(fns) != super().ColumnCount():
                def _arrow248(__unit: None=None) -> str:
                    arg: int = length(fns) or 0
                    arg_1: int = super().ColumnCount() or 0
                    return to_text(printf("The number of field names (%i) must match the number of columns (%i) in the range."))(arg)(arg_1)

                raise Exception(_arrow248(), "fieldNames")

            dict_1: Any = dict([])
            def action(i: int, fn: str) -> None:
                col_i: int = (range_address.FirstAddress.ColumnNumber + i) or 0
                add_to_dict(dict_1, fn, FsTableField(fn, i, FsRangeColumn(FsRangeAddress(FsAddress(range_address.FirstAddress.RowNumber, col_i), FsAddress(range_address.LastAddress.RowNumber, col_i)))))

            iterate_indexed(action, fns)
            return dict_1

        self._fieldNames: Any = dict([]) if (field_names is None) else _arrow249()
        self._uniqueNames: Any = set([])

    @property
    def Name(self, __unit: None=None) -> str:
        this: FsTable = self
        return this._name

    def GetFieldNames(self, cells_collection: FsCellsCollection) -> Any:
        this: FsTable = self
        if equals(this._lastRangeAddress, this.RangeAddress) if ((not equals(this._lastRangeAddress, None)) if (not equals(this._fieldNames, None)) else False) else False:
            return this._fieldNames

        else: 
            this._lastRangeAddress = this.RangeAddress
            return this._fieldNames


    def GetFields(self, cells_collection: FsCellsCollection) -> IEnumerable_1[FsTableField]:
        this: FsTable = self
        def _arrow236(i: int) -> FsTableField:
            return this.GetFieldAt(i, cells_collection)

        return initialize(super().ColumnCount(), _arrow236)

    @property
    def ShowHeaderRow(self, __unit: None=None) -> bool:
        this: FsTable = self
        return this._showHeaderRow

    @ShowHeaderRow.setter
    def ShowHeaderRow(self, show_header_row: bool) -> None:
        this: FsTable = self
        this._showHeaderRow = show_header_row

    def HeadersRow(self, __unit: None=None) -> FsRangeRow:
        this: FsTable = self
        return None if (not this.ShowHeaderRow) else FsRange(super().RangeAddress).FirstRow()

    def TryGetHeaderRow(self, cells_collection: FsCellsCollection) -> FsRow | None:
        this: FsTable = self
        match_value: bool = this.ShowHeaderRow
        if match_value:
            row_index: int = this.RangeAddress.FirstAddress.RowNumber or 0
            return FsRow(FsRangeAddress(FsAddress(row_index, this.RangeAddress.FirstAddress.ColumnNumber), FsAddress(row_index, this.RangeAddress.LastAddress.ColumnNumber)), cells_collection)

        else: 
            return None


    def GetHeaderRow(self, cells_collection: FsCellsCollection) -> FsRow:
        this: FsTable = self
        match_value: FsRow | None = this.TryGetHeaderRow(cells_collection)
        if match_value is None:
            raise Exception(("Error. Unable to get header row for table \"" + this.Name) + "\" as `ShowHeaderRow` is set to `false`.")

        else: 
            return match_value


    def GetColumns(self, cells_collection: FsCellsCollection) -> IEnumerable_1[FsColumn]:
        this: FsTable = self
        def _arrow238(__unit: None=None) -> IEnumerable_1[FsColumn]:
            def _arrow237(i: int) -> IEnumerable_1[FsColumn]:
                return singleton(FsColumn(FsRangeAddress(FsAddress(this.RangeAddress.FirstAddress.RowNumber, i), FsAddress(this.RangeAddress.LastAddress.RowNumber, i)), cells_collection))

            return collect(_arrow237, range_big_int(this.RangeAddress.FirstAddress.ColumnNumber, 1, this.RangeAddress.LastAddress.ColumnNumber))

        return delay(_arrow238)

    def GetRows(self, cells_collection: FsCellsCollection) -> IEnumerable_1[FsRow]:
        this: FsTable = self
        def _arrow240(__unit: None=None) -> IEnumerable_1[FsRow]:
            def _arrow239(i: int) -> IEnumerable_1[FsRow]:
                return singleton(FsRow(FsRangeAddress(FsAddress(i, this.RangeAddress.FirstAddress.ColumnNumber), FsAddress(i, this.RangeAddress.LastAddress.ColumnNumber)), cells_collection))

            return collect(_arrow239, range_big_int(this.RangeAddress.FirstAddress.RowNumber, 1, this.RangeAddress.LastAddress.RowNumber))

        return delay(_arrow240)

    def GetBodyRows(self, cells_collection: FsCellsCollection) -> IEnumerable_1[FsRow]:
        this: FsTable = self
        first_row_index: int = ((this.RangeAddress.FirstAddress.RowNumber + 1) if this.ShowHeaderRow else this.RangeAddress.FirstAddress.RowNumber) or 0
        def _arrow242(__unit: None=None) -> IEnumerable_1[FsRow]:
            def _arrow241(i: int) -> IEnumerable_1[FsRow]:
                return singleton(FsRow(FsRangeAddress(FsAddress(i, this.RangeAddress.FirstAddress.ColumnNumber), FsAddress(i, this.RangeAddress.LastAddress.ColumnNumber)), cells_collection))

            return collect(_arrow241, range_big_int(first_row_index, 1, this.RangeAddress.LastAddress.RowNumber))

        return delay(_arrow242)

    def GetRowAt(self, index: int, cells_collection: FsCellsCollection) -> FsRow:
        this: FsTable = self
        if True if (index <= 0) else (index > super().RowCount()):
            def _arrow243(__unit: None=None) -> str:
                arg: int = super().RowCount() or 0
                return to_text(printf("Row index must be between 1 and %i"))(arg)

            raise Exception(int32_to_string(index), _arrow243())

        row_i: int = ((this.RangeAddress.FirstAddress.RowNumber + index) - 1) or 0
        return FsRow(FsRangeAddress(FsAddress(row_i, this.RangeAddress.FirstAddress.ColumnNumber), FsAddress(row_i, this.RangeAddress.LastAddress.ColumnNumber)), cells_collection)

    def GetBodyRowAt(self, index: int, cells_collection: FsCellsCollection) -> FsRow:
        this: FsTable = self
        return this.GetRowAt(index + 1, cells_collection) if this.ShowHeaderRow else this.GetRowAt(index, cells_collection)

    def RescanRange(self, __unit: None=None) -> None:
        this: FsTable = self
        def reduction(r1: FsRangeAddress, r2: FsRangeAddress) -> FsRangeAddress:
            return r1.Union(r2)

        def mapping(v: FsTableField) -> FsRangeAddress:
            return v.Column.RangeAddress

        range_address: FsRangeAddress = reduce(reduction, map(mapping, this._fieldNames.values()))
        this.RangeAddress = range_address

    @staticmethod
    def rescan_range(table: FsTable) -> FsTable:
        table.RescanRange()
        return table

    def GetUniqueName(self, original_name: str, initial_offset: int, enforce_offset: bool) -> str:
        this: FsTable = self
        name: str = original_name + (int32_to_string(initial_offset) if enforce_offset else "")
        if name in this._uniqueNames:
            i: int = initial_offset or 0
            name = original_name + int32_to_string(i)
            while name in this._uniqueNames:
                i = (i + 1) or 0
                name = original_name + int32_to_string(i)

        ignore(add_to_set(name, this._uniqueNames))
        return name

    @staticmethod
    def get_unique_names(original_name: str, initial_offset: int, enforce_offset: bool, table: FsTable) -> str:
        return table.GetUniqueName(original_name, initial_offset, enforce_offset)

    def InitFields(self, field_names: IEnumerable_1[str]) -> None:
        this: FsTable = self
        def action(i: int, fn: str) -> None:
            table_field: FsTableField = FsTableField(fn, i, FsRangeColumn.from_index(i))
            add_to_dict(this._fieldNames, fn, table_field)

        iterate_indexed(action, field_names)

    @staticmethod
    def init_fields(field_names: IEnumerable_1[str], table: FsTable) -> FsTable:
        table.InitFields(field_names)
        return table

    def AddFields(self, table_fields: IEnumerable_1[FsTableField]) -> None:
        this: FsTable = self
        def action(tf: FsTableField) -> None:
            add_to_dict(this._fieldNames, tf.Name, tf)

        iterate(action, table_fields)

    @staticmethod
    def add_fields(table_fields: IEnumerable_1[FsTableField], table: FsTable) -> FsTable:
        table.AddFields(table_fields)
        return table

    def Field(self, name: str, cells_collection: FsCellsCollection) -> FsTableField:
        this: FsTable = self
        match_value: FsTableField | None = Dictionary_tryGet(name, this._fieldNames)
        if match_value is None:
            def _arrow245(__unit: None=None) -> int:
                def mapping(v: FsTableField) -> int:
                    return v.Index

                s: IEnumerable_1[int] = map(mapping, this._fieldNames.values())
                class ObjectExpr244:
                    @property
                    def Compare(self) -> Callable[[int, int], int]:
                        return compare_primitives

                return 0 if (length(s) == 0) else max(s, ObjectExpr244())

            def _arrow246(__unit: None=None) -> FsRangeAddress:
                offset: int = len(this._fieldNames) or 0
                return FsRangeAddress(FsAddress(this.RangeAddress.FirstAddress.RowNumber, this.RangeAddress.FirstAddress.ColumnNumber + offset), FsAddress(this.RangeAddress.LastAddress.RowNumber, this.RangeAddress.FirstAddress.ColumnNumber + offset))

            new_field: FsTableField = FsTableField(name, _arrow245() + 1, FsRangeColumn(_arrow246()), some(None), some(None))
            if this.ShowHeaderRow:
                value: None = new_field.HeaderCell(cells_collection, True).SetValueAs(name)
                ignore(None)

            add_to_dict(this._fieldNames, name, new_field)
            this.RescanRange()
            return new_field

        else: 
            return match_value


    def GetField(self, name: str, cells_collection: FsCellsCollection) -> FsTableField:
        this: FsTable = self
        name_1: str = replace(name, "\r\n", "\n")
        try: 
            return get_item_from_dict(this.GetFieldNames(cells_collection), name_1)

        except Exception as match_value:
            raise Exception(("The header row doesn\'t contain field name \'" + name_1) + "\'.")


    @staticmethod
    def get_field(name: str, cells_collection: FsCellsCollection, table: FsTable) -> FsTableField:
        return table.GetField(name, cells_collection)

    def GetFieldAt(self, index: int, cells_collection: FsCellsCollection) -> FsTableField:
        this: FsTable = self
        try: 
            def predicate(ftf: FsTableField) -> bool:
                return ftf.Index == index

            return find(predicate, this.GetFieldNames(cells_collection).values())

        except Exception as match_value:
            raise Exception(("FsTableField with index " + str(index)) + " does not exist in the FsTable.")


    def GetFieldIndex(self, name: str, cells_collection: FsCellsCollection) -> int:
        this: FsTable = self
        return this.GetField(name, cells_collection).Index

    def RenameField(self, old_name: str, new_name: str) -> None:
        this: FsTable = self
        match_value: FsTableField | None = Dictionary_tryGet(old_name, this._fieldNames)
        if match_value is None:
            raise Exception("The FsTabelField does not exist in this FsTable", "oldName")

        else: 
            field: FsTableField = match_value
            ignore(remove_from_dict(this._fieldNames, old_name))
            add_to_dict(this._fieldNames, new_name, field)


    @staticmethod
    def rename_field(old_name: str, new_name: str, table: FsTable) -> FsTable:
        table.RenameField(old_name, new_name)
        return table

    def TryGetHeaderCellOfColumnAt(self, cells_collection: FsCellsCollection, col_index: int) -> FsCell | None:
        this: FsTable = self
        fst_row_index: int = this.RangeAddress.FirstAddress.RowNumber or 0
        def predicate(c: FsCell) -> bool:
            return c.RowNumber == fst_row_index

        return try_find(predicate, cells_collection.GetCellsInColumn(col_index))

    @staticmethod
    def try_get_header_cell_of_column_index_at(cells_collection: FsCellsCollection, col_index: int, table: FsTable) -> FsCell | None:
        return table.TryGetHeaderCellOfColumnAt(cells_collection, col_index)

    def TryGetHeaderCellOfColumn(self, cells_collection: FsCellsCollection, column: FsRangeColumn) -> FsCell | None:
        this: FsTable = self
        return this.TryGetHeaderCellOfColumnAt(cells_collection, column.Index)

    @staticmethod
    def try_get_header_cell_of_column(cells_collection: FsCellsCollection, column: FsRangeColumn, table: FsTable) -> FsCell | None:
        return table.TryGetHeaderCellOfColumn(cells_collection, column)

    def GetHeaderCellOfColumnAt(self, cells_collection: FsCellsCollection, col_index: int) -> FsCell:
        this: FsTable = self
        return value_2(this.TryGetHeaderCellOfColumnAt(cells_collection, col_index))

    @staticmethod
    def get_header_cell_of_column_index_at(cells_collection: FsCellsCollection, col_index: int, table: FsTable) -> FsCell:
        return table.GetHeaderCellOfColumnAt(cells_collection, col_index)

    def GetHeaderCellOfColumn(self, cells_collection: FsCellsCollection, column: FsRangeColumn) -> FsCell:
        this: FsTable = self
        return value_2(this.TryGetHeaderCellOfColumn(cells_collection, column))

    @staticmethod
    def get_header_cell_of_column(cells_collection: FsCellsCollection, column: FsRangeColumn, table: FsTable) -> FsCell:
        return table.GetHeaderCellOfColumn(cells_collection, column)

    def GetHeaderCellOfTableField(self, cells_collection: FsCellsCollection, table_field: FsTableField) -> FsCell:
        this: FsTable = self
        return table_field.HeaderCell(cells_collection, this.ShowHeaderRow)

    @staticmethod
    def get_header_cell_of_table_field(cells_collection: FsCellsCollection, table_field: FsTableField, table: FsTable) -> FsCell:
        return table.GetHeaderCellOfTableField(cells_collection, table_field)

    def TryGetHeaderCellOfTableFieldAt(self, cells_collection: FsCellsCollection, table_field_index: int) -> FsCell | None:
        this: FsTable = self
        def chooser(tf: FsTableField) -> FsCell | None:
            if tf.Index == table_field_index:
                return tf.HeaderCell(cells_collection, this.ShowHeaderRow)

            else: 
                return None


        return try_pick(chooser, this._fieldNames.values())

    @staticmethod
    def try_get_header_cell_of_table_field_index_at(cells_collection: FsCellsCollection, table_field_index: int, table: FsTable) -> FsCell | None:
        return table.TryGetHeaderCellOfTableFieldAt(cells_collection, table_field_index)

    def GetHeaderCellOfTableFieldAt(self, cells_collection: FsCellsCollection, table_field_index: int) -> FsCell:
        this: FsTable = self
        return value_2(this.TryGetHeaderCellOfTableFieldAt(cells_collection, table_field_index))

    @staticmethod
    def get_header_cell_of_table_field_index_at(cells_collection: FsCellsCollection, table_field_index: int, table: FsTable) -> FsCell:
        return table.GetHeaderCellOfTableFieldAt(cells_collection, table_field_index)

    def TryGetHeaderCellByFieldName(self, cells_collection: FsCellsCollection, field_name: str) -> FsCell | None:
        this: FsTable = self
        match_value: FsTableField | None = Dictionary_tryGet(field_name, this._fieldNames)
        if match_value is None:
            return None

        else: 
            tf: FsTableField = match_value
            return tf.HeaderCell(cells_collection, this.ShowHeaderRow)


    @staticmethod
    def try_get_header_cell_by_field_name(cells_collection: FsCellsCollection, field_name: str, table: FsTable) -> FsCell | None:
        return table.TryGetHeaderCellByFieldName(cells_collection, field_name)

    def GetDataCellsOfColumnAt(self, cells_collection: FsCellsCollection, col_index: int) -> IEnumerable_1[FsCell]:
        this: FsTable = self
        def chooser(ri: int) -> FsCell | None:
            return cells_collection.TryGetCell(ri, col_index)

        return choose(chooser, to_list(range_big_int(this.RangeAddress.FirstAddress.RowNumber + 1, 1, this.RangeAddress.LastAddress.RowNumber)))

    @staticmethod
    def get_data_cells_of_column_index_at(cells_collection: FsCellsCollection, col_index: int, table: FsTable) -> IEnumerable_1[FsCell]:
        return table.GetDataCellsOfColumnAt(cells_collection, col_index)

    def Copy(self, __unit: None=None) -> FsTable:
        this: FsTable = self
        ra: FsRangeAddress = this.RangeAddress.Copy()
        return FsTable(this.Name, ra, False, this.ShowHeaderRow)

    @staticmethod
    def copy(table: FsTable) -> FsTable:
        return table.Copy()

    @staticmethod
    def validate_for_write(table: FsTable, cells_collection: FsCellsCollection) -> None:
        if length(table.GetBodyRows(cells_collection)) == 0:
            def _arrow247(__unit: None=None) -> str:
                arg: str = table.Name
                return to_text(printf("The table \'%s\' must contain at least one body row."))(arg)

            raise Exception(_arrow247(), "table")


    def RescanFieldNames(self, cells_collection: FsCellsCollection) -> None:
        this: FsTable = self
        if this.ShowHeaderRow:
            old_field_names: Any = this._fieldNames
            this._fieldNames = dict([])
            cell_pos: int = 0
            with get_enumerator(this.GetHeaderRow(cells_collection)) as enumerator:
                while enumerator.System_Collections_IEnumerator_MoveNext():
                    cell: FsCell = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                    name: str = cell.ValueAsString()
                    match_value: FsTableField | None = Dictionary_tryGet(name, old_field_names)
                    if match_value is None:
                        if (name is None) != (name == ""):
                            name = this.GetUniqueName("Column", cell_pos + 1, True)
                            value: None = cell.SetValueAs(name)
                            ignore(None)
                            cell.DataType = DataType(0)

                        if name in this._fieldNames:
                            raise Exception(("The header row contains more than one field name \'" + name) + "\'.")

                        add_to_dict(this._fieldNames, name, FsTableField(name, cell_pos))
                        cell_pos = (cell_pos + 1) or 0

                    else: 
                        table_field: FsTableField = match_value
                        table_field.Index = cell_pos or 0
                        add_to_dict(this._fieldNames, name, table_field)
                        cell_pos = (cell_pos + 1) or 0


        else: 
            col_count: int = super().ColumnCount() or 0
            for i in range(1, col_count + 1, 1):
                def predicate(v: FsTableField) -> bool:
                    return v.Index == (i - 1)

                if not exists(predicate, this._fieldNames.values()):
                    name_1: str = "Column" + int32_to_string(i)
                    add_to_dict(this._fieldNames, name_1, FsTableField(name_1, i - 1))




FsTable_reflection = _expr250

def FsTable__ctor_3B894C50(name: str, range_address: FsRangeAddress, show_totals_row: bool | None=None, show_header_row: bool | None=None, field_names: IEnumerable_1[str] | None=None) -> FsTable:
    return FsTable(name, range_address, show_totals_row, show_header_row, field_names)


__all__ = ["FsTable_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.arc_types import ArcAssay
from ...arctrl_core.comment import (Comment, Remark)
from ...arctrl_core.conversion import (JsonTypes_decomposeTechnologyPlatform, JsonTypes_composeTechnologyPlatform)
from ...arctrl_core.Helper.collections_ import (Option_fromValueWithDefault, ResizeArray_iter)
from ...arctrl_core.Helper.identifier import (create_missing_identifier, Assay_tryIdentifierFromFileName, remove_missing_identifier, Assay_fileNameFromIdentifier)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...fable_library.list import (of_array, FSharpList, length, singleton, initialize, map as map_1, empty, iterate_indexed, cons, reverse)
from ...fable_library.map_util import add_to_dict
from ...fable_library.option import (map, default_arg)
from ...fable_library.seq2 import List_distinct
from ...fable_library.string_ import starts_with_exact
from ...fable_library.types import Array
from ...fable_library.util import (string_hash, IEnumerable_1, IEnumerator)
from .comment import (Comment_fromString, Comment_toString)
from .sparse_table import (SparseTable_GetEmptyComments_3ECCA699, SparseTable__TryGetValueDefault_5BAE6133, SparseTable__TryGetValue_11FD62A8, SparseTable, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1)

labels: FSharpList[str] = of_array(["Identifier", "Title", "Description", "Measurement Type", "Measurement Type Term Accession Number", "Measurement Type Term Source REF", "Technology Type", "Technology Type Term Accession Number", "Technology Type Term Source REF", "Technology Platform", "File Name"])

def from_string(identifier: str | None, title: str | None, description: str | None, measurement_type: str | None, measurement_type_term_source_ref: str | None, measurement_type_term_accession_number: str | None, technology_type: str | None, technology_type_term_source_ref: str | None, technology_type_term_accession_number: str | None, technology_platform: str | None, file_name: str | None, comments: Array[Comment]) -> ArcAssay:
    measurement_type_1: OntologyAnnotation = OntologyAnnotation.create(measurement_type, measurement_type_term_source_ref, measurement_type_term_accession_number)
    technology_type_1: OntologyAnnotation = OntologyAnnotation.create(technology_type, technology_type_term_source_ref, technology_type_term_accession_number)
    identifier_4: str
    if identifier is None:
        if file_name is None:
            identifier_4 = create_missing_identifier()

        else: 
            match_value: str | None = Assay_tryIdentifierFromFileName(file_name)
            identifier_4 = match_value if (match_value is not None) else create_missing_identifier()


    else: 
        identifier_4 = identifier

    measurement_type_2: OntologyAnnotation | None = Option_fromValueWithDefault(OntologyAnnotation(), measurement_type_1)
    technology_type_2: OntologyAnnotation | None = Option_fromValueWithDefault(OntologyAnnotation(), technology_type_1)
    def _arrow3464(name: str, identifier: Any=identifier, title: Any=title, description: Any=description, measurement_type: Any=measurement_type, measurement_type_term_source_ref: Any=measurement_type_term_source_ref, measurement_type_term_accession_number: Any=measurement_type_term_accession_number, technology_type: Any=technology_type, technology_type_term_source_ref: Any=technology_type_term_source_ref, technology_type_term_accession_number: Any=technology_type_term_accession_number, technology_platform: Any=technology_platform, file_name: Any=file_name, comments: Any=comments) -> OntologyAnnotation:
        return JsonTypes_decomposeTechnologyPlatform(name)

    technology_platform_1: OntologyAnnotation | None = map(_arrow3464, technology_platform)
    return ArcAssay.make(identifier_4, title, description, measurement_type_2, technology_type_2, technology_platform_1, [], None, [], comments)


def from_sparse_table(matrix: SparseTable) -> FSharpList[ArcAssay]:
    if (length(matrix.CommentKeys) != 0) if (matrix.ColumnCount == 0) else False:
        comments: Array[Comment] = SparseTable_GetEmptyComments_3ECCA699(matrix)
        return singleton(ArcAssay.create(create_missing_identifier(), None, None, None, None, None, None, None, None, comments))

    else: 
        def _arrow3465(i: int, matrix: Any=matrix) -> ArcAssay:
            def mapping(k: str) -> Comment:
                return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, i)))

            comments_1: Array[Comment] = list(map_1(mapping, matrix.CommentKeys))
            return from_string(SparseTable__TryGetValue_11FD62A8(matrix, ("Identifier", i)), SparseTable__TryGetValue_11FD62A8(matrix, ("Title", i)), SparseTable__TryGetValue_11FD62A8(matrix, ("Description", i)), SparseTable__TryGetValue_11FD62A8(matrix, ("Measurement Type", i)), SparseTable__TryGetValue_11FD62A8(matrix, ("Measurement Type Term Source REF", i)), SparseTable__TryGetValue_11FD62A8(matrix, ("Measurement Type Term Accession Number", i)), SparseTable__TryGetValue_11FD62A8(matrix, ("Technology Type", i)), SparseTable__TryGetValue_11FD62A8(matrix, ("Technology Type Term Source REF", i)), SparseTable__TryGetValue_11FD62A8(matrix, ("Technology Type Term Accession Number", i)), SparseTable__TryGetValue_11FD62A8(matrix, ("Technology Platform", i)), SparseTable__TryGetValue_11FD62A8(matrix, ("File Name", i)), comments_1)

        return initialize(matrix.ColumnCount, _arrow3465)



def to_sparse_table(assays: FSharpList[ArcAssay]) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, labels, None, length(assays) + 1)
    comment_keys: FSharpList[str] = empty()
    def action(i: int, a: ArcAssay, assays: Any=assays) -> None:
        processed_file_name: str = remove_missing_identifier(a.Identifier) if starts_with_exact(a.Identifier, "MISSING_IDENTIFIER_") else Assay_fileNameFromIdentifier(a.Identifier)
        i_1: int = (i + 1) or 0
        mt_1: dict[str, Any]
        mt: OntologyAnnotation = default_arg(a.MeasurementType, OntologyAnnotation())
        mt_1 = OntologyAnnotation.to_string_object(mt, True)
        tt_1: dict[str, Any]
        tt: OntologyAnnotation = default_arg(a.TechnologyType, OntologyAnnotation())
        tt_1 = OntologyAnnotation.to_string_object(tt, True)
        add_to_dict(matrix.Matrix, ("Identifier", i_1), remove_missing_identifier(a.Identifier))
        add_to_dict(matrix.Matrix, ("Title", i_1), default_arg(a.Title, ""))
        add_to_dict(matrix.Matrix, ("Description", i_1), default_arg(a.Description, ""))
        add_to_dict(matrix.Matrix, ("Measurement Type", i_1), mt_1["TermName"])
        add_to_dict(matrix.Matrix, ("Measurement Type Term Accession Number", i_1), mt_1["TermAccessionNumber"])
        add_to_dict(matrix.Matrix, ("Measurement Type Term Source REF", i_1), mt_1["TermSourceREF"])
        add_to_dict(matrix.Matrix, ("Technology Type", i_1), tt_1["TermName"])
        add_to_dict(matrix.Matrix, ("Technology Type Term Accession Number", i_1), tt_1["TermAccessionNumber"])
        add_to_dict(matrix.Matrix, ("Technology Type Term Source REF", i_1), tt_1["TermSourceREF"])
        def mapping(tp: OntologyAnnotation, i: Any=i, a: Any=a) -> str:
            return JsonTypes_composeTechnologyPlatform(tp)

        add_to_dict(matrix.Matrix, ("Technology Platform", i_1), default_arg(map(mapping, a.TechnologyPlatform), ""))
        add_to_dict(matrix.Matrix, ("File Name", i_1), processed_file_name)
        def f(comment: Comment, i: Any=i, a: Any=a) -> None:
            nonlocal comment_keys
            pattern_input: tuple[str, str] = Comment_toString(comment)
            n: str = pattern_input[0]
            comment_keys = cons(n, comment_keys)
            add_to_dict(matrix.Matrix, (n, i_1), pattern_input[1])

        ResizeArray_iter(f, a.Comments)

    iterate_indexed(action, assays)
    class ObjectExpr3467:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3466(x: str, y: str) -> bool:
                return x == y

            return _arrow3466

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3467())), matrix.ColumnCount)


def from_rows(prefix: str | None, line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], FSharpList[ArcAssay]]:
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, labels, line_number, prefix)
    return (tupled_arg[0], tupled_arg[1], tupled_arg[2], from_sparse_table(tupled_arg[3]))


def to_rows(prefix: str | None, assays: FSharpList[ArcAssay]) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    return SparseTable_ToRows_759CAFC1(to_sparse_table(assays), prefix)


__all__ = ["labels", "from_string", "from_sparse_table", "to_sparse_table", "from_rows", "to_rows"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.data import Data
from ...arctrl_core.Process.material import Material
from ...arctrl_core.Process.process_input import ProcessInput
from ...arctrl_core.Process.sample import Sample
from ...arctrl_core.Process.source import Source
from ...fable_library.list import of_array
from ...thoth_json_core.decode import (one_of, map)
from ...thoth_json_core.types import (IEncodable, Decoder_1)
from ..data import (ROCrate_encoder as ROCrate_encoder_2, ROCrate_decoder as ROCrate_decoder_3, ISAJson_encoder as ISAJson_encoder_2, ISAJson_decoder as ISAJson_decoder_3)
from .material import (ROCrate_encoder as ROCrate_encoder_3, ROCrate_decoder as ROCrate_decoder_4, ISAJson_encoder as ISAJson_encoder_3, ISAJson_decoder as ISAJson_decoder_4)
from .sample import (ROCrate_encoder as ROCrate_encoder_1, ROCrate_decoder as ROCrate_decoder_1, ISAJson_encoder as ISAJson_encoder_1, ISAJson_decoder as ISAJson_decoder_2)
from .source import (ROCrate_encoder as ROCrate_encoder_4, ROCrate_decoder as ROCrate_decoder_2, ISAJson_encoder as ISAJson_encoder_4, ISAJson_decoder as ISAJson_decoder_1)

def ROCrate_encoder(value: ProcessInput) -> IEncodable:
    if value.tag == 1:
        return ROCrate_encoder_1(value.fields[0])

    elif value.tag == 2:
        return ROCrate_encoder_2(value.fields[0])

    elif value.tag == 3:
        return ROCrate_encoder_3(value.fields[0])

    else: 
        return ROCrate_encoder_4(value.fields[0])



def _arrow2631(Item: Sample) -> ProcessInput:
    return ProcessInput(1, Item)


def _arrow2632(Item_1: Source) -> ProcessInput:
    return ProcessInput(0, Item_1)


def _arrow2633(Item_2: Data) -> ProcessInput:
    return ProcessInput(2, Item_2)


def _arrow2634(Item_3: Material) -> ProcessInput:
    return ProcessInput(3, Item_3)


ROCrate_decoder: Decoder_1[ProcessInput] = one_of(of_array([map(_arrow2631, ROCrate_decoder_1), map(_arrow2632, ROCrate_decoder_2), map(_arrow2633, ROCrate_decoder_3), map(_arrow2634, ROCrate_decoder_4)]))

def ISAJson_encoder(id_map: Any | None, value: ProcessInput) -> IEncodable:
    if value.tag == 1:
        return ISAJson_encoder_1(id_map, value.fields[0])

    elif value.tag == 2:
        return ISAJson_encoder_2(id_map, value.fields[0])

    elif value.tag == 3:
        return ISAJson_encoder_3(id_map, value.fields[0])

    else: 
        return ISAJson_encoder_4(id_map, value.fields[0])



def _arrow2637(Item: Source) -> ProcessInput:
    return ProcessInput(0, Item)


def _arrow2639(Item_1: Sample) -> ProcessInput:
    return ProcessInput(1, Item_1)


def _arrow2640(Item_2: Data) -> ProcessInput:
    return ProcessInput(2, Item_2)


def _arrow2641(Item_3: Material) -> ProcessInput:
    return ProcessInput(3, Item_3)


ISAJson_decoder: Decoder_1[ProcessInput] = one_of(of_array([map(_arrow2637, ISAJson_decoder_1), map(_arrow2639, ISAJson_decoder_2), map(_arrow2640, ISAJson_decoder_3), map(_arrow2641, ISAJson_decoder_4)]))

__all__ = ["ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_decoder"]

